# -*- coding: utf-8 -*-

from unittest import TestCase
from six.moves.urllib.parse import urlparse, parse_qsl, unquote_plus
from rwslib.rws_requests import ConfigurableDatasetRequest


class Url(object):
    """
    Taken from: http://stackoverflow.com/questions/5371992/comparing-two-urls-in-python
    A url object that can be compared with other url orbjects
    without regard to the vagaries of encoding, escaping, and ordering
    of parameters in query strings.
    """

    def __init__(self, url):
        parts = urlparse(url)
        _query = frozenset(parse_qsl(parts.query))
        _path = unquote_plus(parts.path)
        parts = parts._replace(query=_query, path=_path)
        self.parts = parts

    def __eq__(self, other):
        return self.parts == other.parts

    def __hash__(self):
        return hash(self.parts)


class TestGenericConfigurableDataset(TestCase):
    def test_url_is_constructed_as_expected(self):
        """Given the arguments to the function, we get the correct URL"""
        t = ConfigurableDatasetRequest('SomeCoolSet',
                                       dataset_format="json",
                                       params=dict(subjectid='45838',
                                                   locale='eng',
                                                   app_instance_uuid='1234'))
        self.assertEqual(Url('datasets/SomeCoolSet.json?subjectid=45838&'
                             'locale=eng&'
                             'app_instance_uuid=1234'), Url(t.url_path()))

    def test_specify_format(self):
        """catch the dataset_format if supplied"""
        t = ConfigurableDatasetRequest('SomeCoolSet',
                                       params=dict(subjectid='45838',
                                                   locale='eng',
                                                   app_instance_uuid='1234'))

        self.assertEqual(Url('datasets/SomeCoolSet?subjectid=45838&'
                             'locale=eng&'
                             'app_instance_uuid=1234'), Url(t.url_path()))
        t = ConfigurableDatasetRequest('SomeCoolSet',
                                       dataset_format="csv",
                                       params=dict(subjectid='45838',
                                                   locale='eng',
                                                   app_instance_uuid='1234'))

        self.assertEqual(Url('datasets/SomeCoolSet.csv?subjectid=45838&'
                             'locale=eng&'
                             'app_instance_uuid=1234'), Url(t.url_path()))

    def test_validate_format(self):
        """validate the dataset_format if required"""

        class WidgetConfigurableDataset(ConfigurableDatasetRequest):
            VALID_DATASET_FORMATS = ('json', 'xml')

        t = WidgetConfigurableDataset('SomeCoolSet',
                                      params=dict(subjectid='45838',
                                                  locale='eng',
                                                  app_instance_uuid='1234'))

        self.assertEqual(Url('datasets/SomeCoolSet?subjectid=45838&'
                             'locale=eng&'
                             'app_instance_uuid=1234'), Url(t.url_path()))

        with self.assertRaises(ValueError) as err:
            t = WidgetConfigurableDataset('SomeCoolSet',
                                          dataset_format="tsv",
                                          params=dict(subjectid='45838',
                                                      locale='eng',
                                                      app_instance_uuid='1234'))

        self.assertEqual("Dataset format tsv is not valid for SomeCoolSet", str(err.exception))



# -*- coding: utf-8 -*-

__author__ = 'anewbigging'

from rwslib.extras.rwscmd import data_scrambler
from rwslib.extras.rwscmd.odmutils import E_ODM, A_ODM
import unittest
import datetime
from lxml import etree
from six import string_types

class TestDuckTyping(unittest.TestCase):
    def setUp(self):
        self.values = {
            0: 'int',
            1: 'int',
            -1: 'int',
            '1': 'int',
            '1.0': 'float',
            1.1: 'float',
            'a': 'string',
            '10 MAR 2016': 'date',
            'MAR 2016': 'date',
            '2016': 'date',
            '10 03 2016': 'date',
            '03 2016': 'date',
            '10/MAR/2016': 'date',
            'MAR/2016': 'date',
            '10/03/2016': 'date',
            '03/2016': 'date',
            '10 31 2016': 'string',  # TODO:  check if this is valid Rave date format
            '9999': 'int',
            '16': 'int',
            'MAR': 'string',
            '20:45:23': 'time',
            '20:45': 'time',
            '10:45:23': 'time',
            '10:45': 'time',
            '10:45:23 AM': 'time',
            '10:45 PM': 'time'
        }


    def test_ducktype(self):
        """Test duck typing integers"""
        for value, expected_type in self.values.items():
            rave_type, _ = data_scrambler.typeof_rave_data(value)
            self.assertEqual(expected_type, rave_type,
                            msg='{0} should be of type {1} not {2}'.format(value, expected_type, rave_type))


class TestBasicScrambling(unittest.TestCase):
    def setUp(self):
        self.scr = data_scrambler.Scramble()

    def test_scramble_int(self):
        """Test scrambling integers"""
        i = self.scr.scramble_int(5)
        self.assertEqual(i, str(int(i)))

    def test_scramble_float(self):
        """Test scrambling floats"""
        i = self.scr.scramble_float(5, 2)
        self.assertIsInstance(float(i), float)
        i = self.scr.scramble_float(5, 0)
        self.assertIsInstance(float(i), float)

    def test_scramble_strings(self):
        """Test scrambling strings"""
        i = self.scr.scramble_string(4)
        self.assertEqual(len(i), 4)
        i = self.scr.scramble_string(200)
        self.assertIsInstance(i, string_types)

    def test_scramble_date(self):
        """Test scrambling dates"""
        dt = self.scr.scramble_date('10 MAR 2016')
        self.assertTrue(datetime.datetime.strptime(dt, '%d %b %Y'))
        dt = self.scr.scramble_date('MAR 2016', '%b %Y')
        self.assertTrue(datetime.datetime.strptime(dt, '%b %Y'))

    def test_scramble_time(self):
        """Test scrambling times"""
        dt = self.scr.scramble_time('18:12:14')
        self.assertTrue(datetime.datetime.strptime(dt, '%H:%M:%S'))
        dt = self.scr.scramble_time( '%H:%M')
        self.assertTrue(datetime.datetime.strptime(dt, '%H:%M'))

class TestDuckTypeScrambling(unittest.TestCase):
    def setUp(self):
        self.scr = data_scrambler.Scramble()

    def test_scramble_int(self):
        """Test scrambling integers"""
        i = self.scr.scramble_value('12345')
        self.assertEqual(i, str(int(i)))

    def test_scramble_float(self):
        """Test scrambling floats"""
        i = self.scr.scramble_value('123.45')
        self.assertIsInstance(float(i), float)
        i = self.scr.scramble_value('12345')
        self.assertIsInstance(float(i), float)

    def test_scramble_strings(self):
        """Test scrambling strings"""
        s = 'asdf'
        i = self.scr.scramble_value(s)
        self.assertEqual(len(s), len(i))
        self.assertNotEqual(s, i)
        s = 'This is a large string to test scrambling of large strings'
        i = self.scr.scramble_value(s)
        self.assertNotEqual(s, i)

    def test_scramble_date(self):
        """Test scrambling dates"""
        dt = self.scr.scramble_value('10 MAR 2016')
        self.assertTrue(datetime.datetime.strptime(dt, '%d %b %Y'))
        dt = self.scr.scramble_value('MAR 2016')
        self.assertTrue(datetime.datetime.strptime(dt, '%b %Y'))

    def test_scramble_time(self):
        """Test scrambling times"""
        dt = self.scr.scramble_value('18:12:14')
        self.assertTrue(datetime.datetime.strptime(dt, '%H:%M:%S'))
        dt = self.scr.scramble_value('18:12')
        self.assertTrue(datetime.datetime.strptime(dt, '%H:%M'))

class TestScramblingWithMetadata(unittest.TestCase):
    def setUp(self):
        metadata = """
<ODM FileType="Snapshot" Granularity="Metadata" CreationDateTime="2016-02-29T13:47:23.654-00:00" FileOID="d460fc96-4f08-445f-89b1-0182e8e810c1" ODMVersion="1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" xmlns="http://www.cdisc.org/ns/odm/v1.3">
  <Study OID="Test">
    <GlobalVariables>
      <StudyName>Test</StudyName>
      <StudyDescription></StudyDescription>
      <ProtocolName>Test</ProtocolName>
    </GlobalVariables>
    <BasicDefinitions/>
    <MetaDataVersion OID="1" Name="Metadata version 1">
      <ItemDef OID="VSDT" Name="VSDT" DataType="date" mdsol:DateTimeFormat="dd MMM yyyy" mdsol:VariableOID="VSDT" mdsol:Active="Yes" mdsol:ControlType="DateTime" mdsol:SourceDocument="Yes" mdsol:SASLabel="Visit Date" mdsol:QueryFutureDate="Yes" mdsol:Visible="Yes" mdsol:QueryNonConformance="Yes" mdsol:CanSetStudyEventDate="Yes" />
      <ItemDef OID="TIME" Name="TIME" DataType="time" mdsol:DateTimeFormat="HH:nn:ss" mdsol:VariableOID="TIME" mdsol:Active="Yes" mdsol:ControlType="DateTime" mdsol:SourceDocument="Yes" mdsol:SASLabel="Time" mdsol:QueryFutureDate="Yes" mdsol:Visible="Yes" mdsol:QueryNonConformance="Yes" mdsol:CanSetStudyEventDate="No" />
      <ItemDef OID="VSNUM" Name="VSNUM" DataType="integer" mdsol:VariableOID="VSNUM" Length="2" mdsol:Active="Yes" mdsol:ControlType="DropDownList" mdsol:SourceDocument="Yes" mdsol:SASLabel="Follow Up Visit" mdsol:Visible="Yes" mdsol:QueryNonConformance="Yes" />
      <ItemDef OID="SAE" Name="SAE" DataType="text" mdsol:VariableOID="SAE" Length="200" mdsol:Active="Yes" mdsol:ControlType="LongText" mdsol:SourceDocument="Yes" mdsol:SASLabel="eSAE Desription" mdsol:Visible="Yes" mdsol:QueryNonConformance="Yes" />
      <ItemDef OID="YN" Name="YN" DataType="integer" mdsol:VariableOID="YN" Length="1" mdsol:Active="Yes" mdsol:ControlType="DropDownList" mdsol:SourceDocument="Yes" mdsol:SASLabel="Subject Received Dose" mdsol:Visible="Yes" mdsol:QueryNonConformance="Yes">
        <CodeListRef CodeListOID="YES_NO_UNKNOWN" />
      </ItemDef>
      <CodeList OID="YES_NO_UNKNOWN" Name="YES_NO_UNKNOWN" DataType="integer">
        <CodeListItem CodedValue="0" mdsol:OrderNumber="1" />
        <CodeListItem CodedValue="1" mdsol:OrderNumber="2" />
        <CodeListItem CodedValue="97" mdsol:OrderNumber="3" />
      </CodeList>
    </MetaDataVersion>
  </Study>
</ODM>
"""
        self.scr = data_scrambler.Scramble(metadata=metadata)

    def test_scramble_item_data(self):
        """Test scrambling using metadata"""
        dt = self.scr.scramble_itemdata('VSDT', '')
        self.assertTrue(datetime.datetime.strptime(dt, '%d %b %Y'))

        tm = self.scr.scramble_itemdata('TIME', '')
        self.assertTrue(datetime.datetime.strptime(tm, '%H:%M:%S'))

        st = self.scr.scramble_itemdata('SAE', '')
        self.assertNotEqual(str, st)

        cd = self.scr.scramble_itemdata('YN', '')
        self.assertIn(cd, ['0','1','97'])


    def test_fill_empty(self):
        """Test filling empty values in ODM document"""
        odm = """
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="c3f15f2d-eb69-42e6-bed4-811bff27ebf9" CreationDateTime="2016-03-02T09:27:14.000-00:00">
  <ClinicalData StudyOID="Test(Prod)" MetaDataVersionOID="1">
    <SubjectData SubjectKey="9e15f698-327e-4e9c-8ed5-8be9b27b67b0" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="0100-90005">
      <SiteRef LocationOID="0100"/>
      <StudyEventData StudyEventOID="SCRN">
        <FormData FormOID="FORM1" FormRepeatKey="1">
          <ItemGroupData ItemGroupOID="FORM1">
            <ItemData ItemOID="YN" Value=""/>
          </ItemGroupData>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>
</ODM>
"""

        output = etree.fromstring(self.scr.fill_empty(None, odm))
        path = ".//{0}[@{1}='{2}']".format(E_ODM.ITEM_DATA.value, A_ODM.ITEM_OID.value, 'YN')
        elem = output.find(path)
        self.assertIn(elem.get(A_ODM.VALUE.value), ['0', '1', '97'])


    def test_fill_empty_remove_values(self):
        """Test filling empty values in ODM document"""
        odm = """
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="c3f15f2d-eb69-42e6-bed4-811bff27ebf9" CreationDateTime="2016-03-02T09:27:14.000-00:00">
  <ClinicalData StudyOID="Test(Prod)" MetaDataVersionOID="1">
    <SubjectData SubjectKey="9e15f698-327e-4e9c-8ed5-8be9b27b67b0" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="0100-90005">
      <SiteRef LocationOID="0100"/>
      <StudyEventData StudyEventOID="SCRN">
        <FormData FormOID="FORM1" FormRepeatKey="1">
          <ItemGroupData ItemGroupOID="FORM1">
            <ItemData ItemOID="YN" Value="1"/>
          </ItemGroupData>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>
</ODM>
"""
        output = etree.fromstring(self.scr.fill_empty(None, odm))

        for el in [E_ODM.ITEM_DATA, E_ODM.ITEM_GROUP_DATA, E_ODM.FORM_DATA, E_ODM.STUDY_EVENT_DATA]:
            path = ".//{0}".format(el.value)
            self.assertIsNone(output.find(path))


    def test_fill_empty_remove_values_ny(self):
        """Test filling empty values in ODM document with OID"""
        odm = """
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="c3f15f2d-eb69-42e6-bed4-811bff27ebf9" CreationDateTime="2016-03-02T09:27:14.000-00:00">
  <ClinicalData StudyOID="Test(Prod)" MetaDataVersionOID="1">
    <SubjectData SubjectKey="9e15f698-327e-4e9c-8ed5-8be9b27b67b0" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="0100-90005">
      <SiteRef LocationOID="0100"/>
      <StudyEventData StudyEventOID="SCRN">
        <FormData FormOID="FORM1" FormRepeatKey="1">
          <ItemGroupData ItemGroupOID="FORM1">
            <ItemData ItemOID="YN" Value=""/>
          </ItemGroupData>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>
</ODM>
"""
        fixed_values = {}
        fixed_values['YN'] = '3'
        output = etree.fromstring(self.scr.fill_empty(fixed_values, odm))
        path = ".//{0}[@{1}='{2}']".format(E_ODM.ITEM_DATA.value, A_ODM.ITEM_OID.value, 'YN')
        elem = output.find(path)
        self.assertEqual(elem.get(A_ODM.VALUE.value), '3')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
__author__ = 'anewbigging'

from lxml import etree
from enum import Enum

# Constants
ODM_NS = '{http://www.cdisc.org/ns/odm/v1.3}'
MDSOL_NS = '{http://www.mdsol.com/ns/odm/metadata}'


# Some constant-making helpers
def odm(value):
    """Value prefix with ODM Namespace"""
    return "{0}{1}".format(ODM_NS, value)

def mdsol(value):
    """Value prefix with mdsol Namespace"""
    return "{0}{1}".format(MDSOL_NS, value)


class E_ODM(Enum):
    """
    Defines ODM Elements
    """
    CLINICAL_DATA = odm('ClinicalData')
    SUBJECT_DATA = odm('SubjectData')
    STUDY_EVENT_DATA = odm('StudyEventData')
    FORM_DATA = odm('FormData')
    ITEM_GROUP_DATA = odm('ItemGroupData')
    ITEM_DATA = odm('ItemData')
    METADATA_VERSION = odm('MetaDataVersion')

    USER_REF = odm('UserRef')
    SOURCE_ID = odm('SourceID')
    DATE_TIME_STAMP_ = odm('DateTimeStamp')
    REASON_FOR_CHANGE = odm('ReasonForChange')
    LOCATION_REF = odm('LocationRef')
    QUERY = mdsol('Query')
    PROTOCOL_DEVIATION = mdsol('ProtocolDeviation')
    REVIEW = mdsol('Review')
    COMMENT = mdsol('Comment')
    SIGNATURE = odm('Signature')
    SIGNATURE_REF = odm('SignatureRef')
    SOURCEID = odm('SourceID')

    ITEM_DEF = odm('ItemDef')
    RANGE_CHECK = odm('RangeCheck')
    CODELIST_REF = odm('CodeListRef')
    CODELIST = odm('CodeList')
    CODELIST_ITEM = odm('CodeListItem')
    ENUMERATED_ITEM = odm('EnumeratedItem')


class A_ODM(Enum):
    """
    Defines ODM Attributes
    """
    AUDIT_SUBCATEGORY_NAME = mdsol('AuditSubCategoryName')
    METADATA_VERSION_OID = 'MetaDataVersionOID'
    STUDY_OID = 'StudyOID'
    TRANSACTION_TYPE = 'TransactionType'
    SUBJECT_NAME = mdsol('SubjectName')
    SUBJECT_KEY = 'SubjectKey'
    USER_OID = 'UserOID'
    LOCATION_OID = 'LocationOID'
    ITEM_OID = 'ItemOID'
    VALUE = 'Value'
    STUDYEVENT_OID = 'StudyEventOID'
    STUDYEVENT_REPEAT_KEY = 'StudyEventRepeatKey'
    FORM_OID = 'FormOID'
    FORM_REPEAT_KEY = 'FormRepeatKey'
    ITEMGROUP_OID = 'ItemGroupOID'
    ITEMGROUP_REPEAT_KEY = 'ItemGroupRepeatKey'
    QUERY_REPEAT_KEY = 'QueryRepeatKey'
    STATUS = 'Status'
    RECIPIENT = 'Recipient'
    RESPONSE = 'Response'
    FREEZE = mdsol('Freeze')
    VERIFY = mdsol('Verify')
    LOCK = mdsol('Lock')
    SUBJECT_STATUS = mdsol('Status')
    PROTCOL_DEVIATION_REPEAT_KEY = 'ProtocolDeviationRepeatKey'
    CLASS = 'Class'  # PV
    CODE = 'Code'  # PV
    REVIEWED = 'Reviewed' #Reviews
    GROUP_NAME = 'GroupName'
    COMMENT_REPEAT_KEY = 'CommentRepeatKey'
    INSTANCE_NAME = mdsol('InstanceName')
    INSTANCE_OVERDUE = mdsol('InstanceOverdue')
    DATAPAGE_NAME = mdsol('DataPageName')
    SIGNATURE_OID = 'SignatureOID'

    OID = 'OID'
    DATATYPE = 'DataType'
    LENGTH = 'Length'
    SIGNIFICANT_DIGITS = 'SignficantDigits'
    CODELIST_OID = 'CodeListOID'
    CODED_VALUE = 'CodedValue'
    DATETIME_FORMAT = mdsol('DateTimeFormat')


def xml_pretty_print(str):
    parser = etree.XMLParser(remove_blank_text=True)
    xml = etree.fromstring(str, parser)
    return etree.tostring(xml, pretty_print=True)




import requests

from .rws_requests import RWSRequest, make_url
from .rwsobjects import RWSException, RWSError, RWSErrorResponse, RWSPostErrorResponse

import time

# -------------------------------------------------------------------------------------------------------
# Classes


class AuthorizationException(Exception):
    """Raised if a request requires authorization but no authorization header is provided"""

    pass


class RWSConnection(object):
    """A connection to RWS"""

    def __init__(
        self,
        domain,
        username=None,
        password=None,
        auth=None,
        virtual_dir="RaveWebServices",
    ):
        """
        Create a connection to Rave

        :param str domain: Rave URL Name
        :param str username: Rave User Login
        :param str password: Rave User password
        :param str auth: Authentication tuple (usually something like `(username, password)`
        :param str virtual_dir: Name of the Rave Web Services prefix (usually `RaveWebServices`, but can be customised)

        .. note::
            If the `domain` does not start with http then it is assumed to be the name of the Medidata
            url and https:// will be added as a prefix and .mdsol.com will be added as a postfix.

            * innovate => https://innovate.mdsol.com
            * http://mytest => http:/mytest

        """

        if domain.lower().startswith("http"):
            self.domain = domain
        else:
            self.domain = "https://%s.mdsol.com" % domain

        self.auth = None
        if auth is not None:
            self.auth = auth
        elif username is not None and password is not None:
            # Make a basic auth
            self.auth = (username, password)

        self.base_url = make_url(self.domain, virtual_dir)

        # Keep track of results of last request so users can get if they need.
        self.last_result = None

        # Time taken to process last request
        self.request_time = None

    def send_request(self, request_object, timeout=None, retries=1, preserve_last_response=True, **kwargs):
        """Send request to RWS endpoint. The request object passed provides the URL endpoint and the HTTP method.
           Takes the text response from RWS and allows the request object to modify it for return. This allows the request
           object to return text, an XML document object, a CSV file or anything else that can be generated from the text
           response from RWS.
           A timeout, in seconds, can be optionally passed into send_request.
           Preserving response in class can be disabled by setting `preserve_last_response=False`.
        """
        if not isinstance(request_object, RWSRequest):
            raise ValueError("Request object must be a subclass of RWSRequest")

        # Construct a URL from the object and make a call
        full_url = make_url(self.base_url, request_object.url_path())
        if request_object.requires_authorization:
            kwargs["auth"] = self.auth
            # TODO: Look at different connect and read timeouts?
            kwargs["timeout"] = timeout
            kwargs.update(request_object.args())

        # Explicit use of requests library here. Could alter in future to inject library to use in case
        # requests not available.

        # Get a session that allows us to customize HTTP requests
        session = requests.Session()

        # Mount a custom adapter that retries failed connections for HTTP and HTTPS requests.
        for scheme in ["http://", "https://"]:
            session.mount(scheme, requests.adapters.HTTPAdapter(max_retries=retries))

        action = {"GET": session.get, "POST": session.post}[request_object.method]

        start_time = time.time()

        try:
            r = action(full_url, **kwargs)  # type: requests.models.Response
        except requests.exceptions.ConnectTimeout:
            raise RWSException(
                "Server Connection Timeout",
                "Connection timeout for {}".format(full_url),
            )
        except requests.exceptions.ReadTimeout:
            raise RWSException(
                "Server Read Timeout", "Read timeout for {}".format(full_url)
            )

        self.request_time = time.time() - start_time
        if preserve_last_response:
            self.last_result = r  # see also r.elapsed for timedelta object.

        if r.status_code in [400, 404]:
            # Is it a RWS response?
            if r.text.startswith("<Response"):
                error = RWSErrorResponse(r.text) if request_object.method == "GET" else RWSPostErrorResponse(r.text)
                raise RWSException(error.errordescription, error)
            elif "<html" in r.text:
                raise RWSException("IIS Error", r.text)
            else:
                error = RWSError(r.text)
            raise RWSException(error.errordescription, error)

        elif r.status_code == 500:
            raise RWSException("Server Error (500)", r.text)

        elif r.status_code == 401:
            # Either you didn't supply auth header and it was required OR your credentials were wrong
            # RWS handles each differently

            # You didn't supply auth (text response from RWS)
            if r.text == "Authorization Header not provided":
                raise AuthorizationException(r.text)

            if "HTTP Error 401.0 - Unauthorized" in r.text:
                raise RWSException("Unauthorized.", r.text)

            # Check if the content_type is text/xml.  Use startswith
            # in case the charset is also specified:
            #  content-type: text/xml; charset=utf-8
            if r.headers.get("content-type").startswith("text/xml"):
                # XML response
                if r.text.startswith("<Response"):
                    error = RWSErrorResponse(r.text) if request_object.method == "GET" else RWSPostErrorResponse(r.text)
                elif "ODM" in r.text:
                    error = RWSError(r.text)
            else:
                # There was some problem with your credentials (XML response from RWS)
                error = RWSErrorResponse(r.text)
            raise RWSException(error.errordescription, error)

        # Catch all.
        if r.status_code != 200:
            if "<" in r.text:
                # XML like
                if r.text.strip().startswith("<Response"):
                    error = RWSErrorResponse(r.text) if request_object.method == "GET" else RWSPostErrorResponse(r.text)
                elif "ODM" in r.text:
                    error = RWSError(r.text)
                else:
                    # IIS error page as an example
                    raise RWSException(
                        "Unexpected Status Code ({0.status_code})".format(r), r.text
                    )
            else:
                # not XML like, better to be safe than blow up
                # example response: 'HTTP 503 Service Temporarily Unavailable'
                raise RWSException(
                    "Unexpected Status Code ({0.status_code})".format(r), r.text
                )
            raise RWSException(error.errordescription, error)

        return request_object.result(r)



# -*- coding: utf-8 -*-
import six

__author__ = 'isparks'

from lxml import etree
import datetime
from rwslib.extras.audit_event.context import (Context, Subject, Event,
                                                Form, ItemGroup, Item,
                                                Query, Review, Comment,
                                                ProtocolDeviation)

# Constants
ODM_NS = '{http://www.cdisc.org/ns/odm/v1.3}'
MDSOL_NS = '{http://www.mdsol.com/ns/odm/metadata}'


# Some constant-making helpers
def odm(value):
    """Value prefix with ODM Namespace"""
    return "{0}{1}".format(ODM_NS, value)


def mdsol(value):
    """Value prefix with mdsol Namespace"""
    return "{0}{1}".format(MDSOL_NS, value)


def yes_no_none(value):
    """Convert Yes/No/None to True/False/None"""
    if not value:
        return None

    # Yes = True, anything else false
    return value.lower() == 'yes'


def make_int(value, missing=-1):
    """Convert string value to long, '' to missing"""
    if isinstance(value, six.string_types):
        if not value.strip():
            return missing
    elif value is None:
        return missing
    return int(value)

# Defaults
DEFAULT_TRANSACTION_TYPE = u'Upsert'

# Attributes
A_AUDIT_SUBCATEGORY_NAME = mdsol('AuditSubCategoryName')
A_METADATA_VERSION_OID = 'MetaDataVersionOID'
A_STUDY_OID = 'StudyOID'
A_TRANSACTION_TYPE = 'TransactionType'
A_SUBJECT_NAME = mdsol('SubjectName')
A_SUBJECT_KEY = 'SubjectKey'
A_USER_OID = 'UserOID'
A_LOCATION_OID = 'LocationOID'
A_ITEM_OID = 'ItemOID'
A_VALUE = 'Value'
A_STUDYEVENT_OID = 'StudyEventOID'
A_STUDYEVENT_REPEAT_KEY = 'StudyEventRepeatKey'
A_RECORD_ID = mdsol('RecordId')
A_FORM_OID = 'FormOID'
A_FORM_REPEAT_KEY = 'FormRepeatKey'
A_ITEMGROUP_OID = 'ItemGroupOID'
A_ITEMGROUP_REPEAT_KEY = 'ItemGroupRepeatKey'
A_QUERY_REPEAT_KEY = 'QueryRepeatKey'
A_STATUS = 'Status'
A_RECIPIENT = 'Recipient'
A_RESPONSE = 'Response'
A_FREEZE = mdsol('Freeze')
A_VERIFY = mdsol('Verify')
A_LOCK = mdsol('Lock')
A_SUBJECT_STATUS = mdsol('Status')
A_PROTCOL_DEVIATION_REPEAT_KEY = 'ProtocolDeviationRepeatKey'
A_CLASS = 'Class'  # PV
A_CODE = 'Code'  # PV
A_REVIEWED = 'Reviewed'  # Reviews
A_GROUP_NAME = 'GroupName'
A_COMMENT_REPEAT_KEY = 'CommentRepeatKey'
A_INSTANCE_ID = mdsol('InstanceId')
A_INSTANCE_NAME = mdsol('InstanceName')
A_INSTANCE_OVERDUE = mdsol('InstanceOverdue')
A_DATAPAGE_NAME = mdsol('DataPageName')
A_DATAPAGE_ID = mdsol('DataPageId')
A_SIGNATURE_OID = 'SignatureOID'
A_SPECIFY_VALUE = mdsol('SpecifyValue')
A_SIGNATURE_BROKEN = mdsol('SignatureBroken')



# Elements
E_CLINICAL_DATA = odm('ClinicalData')
E_SUBJECT_DATA = odm('SubjectData')
E_USER_REF = odm('UserRef')
E_SOURCE_ID = odm('SourceID')
E_DATE_TIME_STAMP_ = odm('DateTimeStamp')
E_REASON_FOR_CHANGE = odm('ReasonForChange')
E_LOCATION_REF = odm('LocationRef')
E_STUDY_EVENT_DATA = odm('StudyEventData')
E_FORM_DATA = odm('FormData')
E_ITEM_GROUP_DATA = odm('ItemGroupData')
E_ITEM_DATA = odm('ItemData')
E_QUERY = mdsol('Query')
E_PROTOCOL_DEVIATION = mdsol('ProtocolDeviation')
E_REVIEW = mdsol('Review')
E_COMMENT = mdsol('Comment')
E_SIGNATURE = odm('Signature')
E_SIGNATURE_REF = odm('SignatureRef')

# Parser States
STATE_NONE = 0
STATE_SOURCE_ID = 1
STATE_DATETIME = 2
STATE_REASON_FOR_CHANGE = 3

# Signature elements have some of the same elements as Audits. Tells us which we are collecting
AUDIT_REF_STATE = 0
SIGNATURE_REF_STATE = 1


class ODMTargetParser(object):
    """A SAX-style lxml Target parser class"""

    def __init__(self, handler):

        # Handler, object that deals with emitting entries etc
        self.handler = handler

        # Context holds the current set of values we are building up, ready to emit
        self.context = None

        # State controls when we are looking for element text content what attribute to put it in
        self.state = STATE_NONE

        # Count of elements processed
        self.count = 0

        self.ref_state = AUDIT_REF_STATE

    def emit(self):
        """We are finished processing one element. Emit it"""

        self.count += 1

        # event_name = 'on_{0}'.format(self.context.subcategory.lower())
        event_name = self.context.subcategory
        if hasattr(self.handler, event_name):
            getattr(self.handler, event_name)(self.context)
        elif hasattr(self.handler, 'default'):
            self.handler.default(self.context)

    def start(self, tag, attrib):
        """On start of element tag"""
        if tag == E_CLINICAL_DATA:
            self.ref_state = AUDIT_REF_STATE
            self.context = Context(attrib[A_STUDY_OID],
                                   attrib[A_AUDIT_SUBCATEGORY_NAME],
                                   int(attrib[A_METADATA_VERSION_OID]))

        elif tag == E_SUBJECT_DATA:
            self.context.subject = Subject(
                attrib.get(A_SUBJECT_KEY),
                attrib.get(A_SUBJECT_NAME),
                attrib.get(A_SUBJECT_STATUS),
                attrib.get(A_TRANSACTION_TYPE, DEFAULT_TRANSACTION_TYPE),
            )
        elif tag == E_USER_REF:
            # Set the Signature or audit-record value depending on state
            self.get_parent_element().user_oid = attrib.get(A_USER_OID)

        elif tag == E_SOURCE_ID:
            self.state = STATE_SOURCE_ID

        elif tag == E_DATE_TIME_STAMP_:
            self.state = STATE_DATETIME

        elif tag == E_REASON_FOR_CHANGE:
            self.state = STATE_REASON_FOR_CHANGE

        elif tag == E_LOCATION_REF:
            # Set the Signature or audit-record value depending on state
            self.get_parent_element().location_oid = attrib.get(A_LOCATION_OID)

        elif tag == E_STUDY_EVENT_DATA:
            self.context.event = Event(
                attrib.get(A_STUDYEVENT_OID),
                attrib.get(A_STUDYEVENT_REPEAT_KEY),
                attrib.get(A_TRANSACTION_TYPE),
                attrib.get(A_INSTANCE_NAME),
                attrib.get(A_INSTANCE_OVERDUE),
                make_int(attrib.get(A_INSTANCE_ID), -1)
            )

        elif tag == E_FORM_DATA:
            self.context.form = Form(
                attrib.get(A_FORM_OID),
                int(attrib.get(A_FORM_REPEAT_KEY, 0)),
                attrib.get(A_TRANSACTION_TYPE),
                attrib.get(A_DATAPAGE_NAME),
                make_int(attrib.get(A_DATAPAGE_ID, -1)),
            )

        elif tag == E_ITEM_GROUP_DATA:
            self.context.itemgroup = ItemGroup(
                attrib.get(A_ITEMGROUP_OID),
                int(attrib.get(A_ITEMGROUP_REPEAT_KEY, 0)),
                attrib.get(A_TRANSACTION_TYPE),
                make_int(attrib.get(A_RECORD_ID, -1)),
            )

        elif tag == E_ITEM_DATA:
            self.context.item = Item(
                attrib.get(A_ITEM_OID),
                attrib.get(A_VALUE),
                attrib.get(A_SPECIFY_VALUE),
                yes_no_none(attrib.get(A_SIGNATURE_BROKEN)),
                yes_no_none(attrib.get(A_FREEZE)),
                yes_no_none(attrib.get(A_VERIFY)),
                yes_no_none(attrib.get(A_LOCK)),
                attrib.get(A_TRANSACTION_TYPE)
            )

        elif tag == E_QUERY:
            self.context.query = Query(
                make_int(attrib.get(A_QUERY_REPEAT_KEY, -1)),
                attrib.get(A_STATUS),
                attrib.get(A_RESPONSE),
                attrib.get(A_RECIPIENT),
                attrib.get(A_VALUE)  # Optional, depends on status
            )

        elif tag == E_PROTOCOL_DEVIATION:
            self.context.protocol_deviation = ProtocolDeviation(
                make_int(attrib.get(A_PROTCOL_DEVIATION_REPEAT_KEY, -1)),
                attrib.get(A_CODE),
                attrib.get(A_CLASS),
                attrib.get(A_STATUS),
                attrib.get(A_VALUE),
                attrib.get(A_TRANSACTION_TYPE)
            )

        elif tag == E_REVIEW:
            self.context.review = Review(
                attrib.get(A_GROUP_NAME),
                yes_no_none(attrib.get(A_REVIEWED)),
            )
        elif tag == E_COMMENT:
            self.context.comment = Comment(
                attrib.get(A_COMMENT_REPEAT_KEY),
                attrib.get(A_VALUE),
                attrib.get(A_TRANSACTION_TYPE)

            )
        elif tag == E_SIGNATURE:
            self.ref_state = SIGNATURE_REF_STATE

        elif tag == E_SIGNATURE_REF:
            self.context.signature.oid = attrib.get(A_SIGNATURE_OID)

    def end(self, tag):
        """Detect end of element"""
        # Emit the context if we reach the end of the audit section
        if tag == E_CLINICAL_DATA:
            self.emit()

    def get_parent_element(self):
        """Signatures and Audit elements share sub-elements, we need to know which to set attributes on"""
        return {AUDIT_REF_STATE: self.context.audit_record,
                SIGNATURE_REF_STATE: self.context.signature}[self.ref_state]

    def data(self, data):
        """Called for text between tags"""
        if self.state == STATE_SOURCE_ID:
            self.context.audit_record.source_id = int(data)  # Audit ids can be 64 bits
        elif self.state == STATE_DATETIME:
            dt = datetime.datetime.strptime(data, "%Y-%m-%dT%H:%M:%S")
            self.get_parent_element().datetimestamp = dt
        elif self.state == STATE_REASON_FOR_CHANGE:
            self.context.audit_record.reason_for_change = data.strip() or None  # Convert a result of '' to None.
        self.state = STATE_NONE

    def close(self):
        self.context = None
        return self.count


def parse(data, eventer):
    """Parse the XML data, firing events from the eventer"""
    parser = etree.XMLParser(target=ODMTargetParser(eventer))
    return etree.XML(data, parser)  # Returns value of close



# -*- coding: utf-8 -*-
import unittest
from rwslib.extras.audit_event.context import ContextBase


class ContextBaseTestCase(unittest.TestCase):

    def setUp(self):
        pass

    def test_repr(self):
        self.assertEqual(ContextBase().__repr__(), 'ContextBase({})')



# -*- coding: utf-8 -*-
from typing import Optional

from rwslib.builders.common import (
    ODMElement,
    TransactionalElement,
    bool_to_yes_no,
    dt_to_iso8601,
    VALID_ID_CHARS,
)
from rwslib.builders.modm import LastUpdateMixin, MilestoneMixin
from rwslib.builders.metadata import MeasurementUnitRef
from rwslib.builders.constants import ProtocolDeviationStatus, QueryStatusType, LocationOIDType
from rwslib.builders.common import Unset

from collections import OrderedDict
from datetime import datetime
import re


class ClinicalData(ODMElement, LastUpdateMixin):
    """Models the ODM ClinicalData object"""

    def __init__(
        self, projectname, environment, metadata_version_oid="1", annotations=None
    ):
        """
        :param projectname: Name of Project in Medidata Rave
        :param environment: Rave Study Enviroment
        :param metadata_version_oid: MetadataVersion OID
        """
        super(ClinicalData, self).__init__()
        self.projectname = projectname
        self.environment = environment
        self.metadata_version_oid = metadata_version_oid
        #: collection of :class:`SubjectData` for the ClinicalData Element
        self.subject_data = []
        self.annotations = annotations

    def build(self, builder):
        """Build XML by appending to builder"""

        # recently we have had issues with the Prod designation on Study OID
        # it defaults to Prod so we can use that
        _study_oid = self.projectname
        if self.environment.lower() != "prod":
            _study_oid += " (%s)" % self.environment
        params = dict(
            MetaDataVersionOID=str(self.metadata_version_oid),
            StudyOID=_study_oid,
        )

        # mixins
        self.mixin_params(params)

        builder.start("ClinicalData", params)
        # Ask children
        if self.subject_data:
            for subject in self.subject_data:
                subject.build(builder)
        # Add the Annotations
        if self.annotations is not None:
            self.annotations.build(builder)
        builder.end("ClinicalData")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (SubjectData, Annotations)):
            raise ValueError(
                "ClinicalData object can only receive SubjectData or Annotations object"
            )
        self.set_list_attribute(other, SubjectData, "subject_data")
        self.set_single_attribute(other, Annotations, "annotations")
        return other


class SubjectData(TransactionalElement, LastUpdateMixin, MilestoneMixin):
    """Models the ODM SubjectData and ODM SiteRef objects"""

    ALLOWED_TRANSACTION_TYPES = ["Insert", "Update", "Upsert", "Context", "Remove"]

    def __init__(
        self,
        site_location_oid,
        subject_key,
        subject_key_type="SubjectName",
        transaction_type="Update",
        location_oid_type: Optional[LocationOIDType] = None,
    ):
        """
        :param str site_location_oid: :class:`SiteLocation` OID
        :param str subject_key: Value for SubjectKey
        :param str subject_key_type: Specifier as to the type of SubjectKey (either **SubjectName** or **SubjectUUID**)
        :param str transaction_type: Transaction Type for Data (one of **Insert**, **Update**, **Upsert**, **Context**, **Remove**)
        """
        super(self.__class__, self).__init__(transaction_type)
        self.sitelocationoid = site_location_oid
        self.subject_key = subject_key
        self.subject_key_type = subject_key_type
        #: collection of :class:`StudyEventData`
        self.study_events = []
        #: collection of :class:`Annotation`
        self.annotations = []
        #: :class:`AuditRecord` for SubjectData - *Not Supported By Rave*
        self.audit_record = None
        #: :class:`Signature` for SubjectData
        self.signature = None
        #: :class:`SiteRef`
        self.siteref = None
        #: :class:`LocationOIDType`
        self.location_oid_type = location_oid_type

    def build(self, builder):
        """Build XML by appending to builder"""
        params = dict(SubjectKey=self.subject_key)
        params["mdsol:SubjectKeyType"] = self.subject_key_type

        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        # mixins
        self.mixin()
        self.mixin_params(params)

        builder.start("SubjectData", params)

        # Ask children
        if self.audit_record is not None:
            self.audit_record.build(builder)

        if self.siteref:
            self.siteref.build(builder)
        else:
            _siteref = SiteRef(self.sitelocationoid, self.location_oid_type)
            _siteref.build(builder)

        for event in self.study_events:
            event.build(builder)

        if self.signature is not None:
            self.signature.build(builder)

        for annotation in self.annotations:
            annotation.build(builder)

        builder.end("SubjectData")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(
            other, (StudyEventData, AuditRecord, Annotation, Signature, SiteRef)
        ):
            raise ValueError(
                "SubjectData object can only receive StudyEventData, AuditRecord, "
                "Annotation or Signature object"
            )

        self.set_list_attribute(other, Annotation, "annotations")
        self.set_list_attribute(other, StudyEventData, "study_events")
        self.set_single_attribute(other, AuditRecord, "audit_record")
        self.set_single_attribute(other, Signature, "signature")
        self.set_single_attribute(other, SiteRef, "siteref")

        return other


class StudyEventData(TransactionalElement, LastUpdateMixin, MilestoneMixin):
    """Models the ODM StudyEventData object"""

    ALLOWED_TRANSACTION_TYPES = ["Insert", "Update", "Upsert", "Context", "Remove"]

    def __init__(
        self, study_event_oid, transaction_type="Update", study_event_repeat_key=None
    ):
        """
        :param str study_event_oid: :class:`StudyEvent` OID
        :param str transaction_type: Transaction Type for Data (one of **Insert**, **Update**, **Upsert**, **Context**, **Remove**)
        :param int study_event_repeat_key: :attr:`StudyEventRepeatKey` for StudyEventData
        """
        super(self.__class__, self).__init__(transaction_type)
        self.study_event_oid = study_event_oid
        self.study_event_repeat_key = study_event_repeat_key
        #: :class:`FormData` part of Study Event Data
        self.forms = []
        #: :class:`Annotation` for Study Event Data  - *Not Supported by Rave*
        self.annotations = []
        #: :class:`Signature` for Study Event Data
        self.signature = None

    def build(self, builder):
        """Build XML by appending to builder
        :Example:
        <StudyEventData StudyEventOID="SCREENING" StudyEventRepeatKey="1" TransactionType="Update">

        """
        params = dict(StudyEventOID=self.study_event_oid)

        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        if self.study_event_repeat_key is not None:
            params["StudyEventRepeatKey"] = str(self.study_event_repeat_key)

        # mixins
        self.mixin()
        self.mixin_params(params)

        builder.start("StudyEventData", params)

        # Ask children
        for form in self.forms:
            form.build(builder)

        if self.signature is not None:
            self.signature.build(builder)

        for annotation in self.annotations:
            annotation.build(builder)

        builder.end("StudyEventData")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (FormData, Annotation, Signature)):
            raise ValueError(
                "StudyEventData object can only receive FormData, Signature or Annotation objects"
            )
        self.set_list_attribute(other, FormData, "forms")
        self.set_single_attribute(other, Signature, "signature")
        self.set_list_attribute(other, Annotation, "annotations")
        return other


class FormData(TransactionalElement, LastUpdateMixin, MilestoneMixin):
    """Models the ODM FormData object"""

    ALLOWED_TRANSACTION_TYPES = ["Insert", "Update", "Upsert", "Context", "Remove"]

    def __init__(self, formoid, transaction_type=None, form_repeat_key=None, lab_reference=None, lab_type=None):
        """
        :param str formoid: :class:`FormDef` OID
        :param str transaction_type: Transaction Type for Data (one of **Insert**, **Update**, **Upsert**, **Context**, **Remove**)
        :param str form_repeat_key: Repeat Key for FormData
        :param str lab_reference: Name for the Lab with the ranges
        :param LabType lab_type: Type of the Lab (Central, Local)
        """
        super(FormData, self).__init__(transaction_type)
        self.formoid = formoid
        self.form_repeat_key = form_repeat_key
        self.lab_reference = lab_reference
        self.lab_type = lab_type
        self.itemgroups = []
        #: :class:`Signature` for FormData
        self.signature = None  # type: Signature
        #: Collection of :class:`Annotation` for FormData - *Not supported by Rave*
        self.annotations = []  # type: list(Annotation)

    def build(self, builder):
        """Build XML by appending to builder
        :Example:
        <FormData FormOID="MH" TransactionType="Update">

        """
        params = dict(FormOID=self.formoid)

        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        if self.form_repeat_key is not None:
            params["FormRepeatKey"] = str(self.form_repeat_key)

        if self.lab_reference is not None:
            params["mdsol:LaboratoryRef"] = str(self.lab_reference)

        if self.lab_type is not None:
            params["mdsol:LaboratoryType"] = self.lab_type.value

        # mixins
        self.mixin()
        self.mixin_params(params)

        builder.start("FormData", params)

        # Ask children
        for itemgroup in self.itemgroups:
            itemgroup.build(builder, self.formoid)

        if self.signature is not None:
            self.signature.build(builder)

        for annotation in self.annotations:
            annotation.build(builder)

        builder.end("FormData")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (Signature, ItemGroupData, Annotation)):
            raise ValueError(
                "FormData object can only receive ItemGroupData, Signature or Annotation objects (not '{}')".format(
                    other
                )
            )
        self.set_list_attribute(other, ItemGroupData, "itemgroups")
        self.set_list_attribute(other, Annotation, "annotations")
        self.set_single_attribute(other, Signature, "signature")
        return other


class ItemGroupData(TransactionalElement, LastUpdateMixin, MilestoneMixin):
    """
    Models the ODM ItemGroupData object.

    .. note:: No name for the ItemGroupData element is required. This is built automatically by the form.
    """

    ALLOWED_TRANSACTION_TYPES = ["Insert", "Update", "Upsert", "Context", "Remove"]

    def __init__(
        self,
        itemgroupoid=None,
        transaction_type=None,
        item_group_repeat_key=None,
        whole_item_group=False,
        annotations=None,
    ):
        """
        :param str transaction_type: Transaction Type for Data (one of **Insert**, **Update**, **Upsert**, **Context**, **Remove**)
        :param int item_group_repeat_key: RepeatKey for the ItemGroupData
        :param bool whole_item_group: Is this the entire ItemGroupData, or just parts? - *Rave specific attribute*
        :param annotations: Annotation for the ItemGroup - *Not supported by Rave*
        :type annotations: list(Annotation) or Annotation
        """
        super(self.__class__, self).__init__(transaction_type)
        self.item_group_repeat_key = item_group_repeat_key
        self.whole_item_group = whole_item_group
        self.items = OrderedDict()
        self.annotations = []
        if annotations:
            # Add the annotations
            if isinstance(annotations, Annotation):
                self << annotations
            elif isinstance(annotations, list):
                for annotation in annotations:
                    self << annotation
        #: :class:`Signature` for ItemGroupData
        self.signature = None
        self.itemgroupoid = itemgroupoid

    def build(self, builder, formname=None):
        """Build XML by appending to builder"""

        params = dict(ItemGroupOID=formname if formname else self.itemgroupoid)

        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        if self.item_group_repeat_key is not None:
            params["ItemGroupRepeatKey"] = str(
                self.item_group_repeat_key
            )  # may be @context for transaction type upsert or context

        params["mdsol:Submission"] = (
            "WholeItemGroup" if self.whole_item_group else "SpecifiedItemsOnly"
        )

        # mixins
        self.mixin()
        self.mixin_params(params)

        builder.start("ItemGroupData", params)

        # Ask children
        for item in self.items.values():
            item.build(builder)

        # Add annotations
        for annotation in self.annotations:
            annotation.build(builder)

        # Add the signature if it exists
        if self.signature is not None:
            self.signature.build(builder)
        builder.end("ItemGroupData")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (ItemData, Annotation, Signature)):
            raise ValueError(
                "ItemGroupData object can only receive ItemData, Signature or Annotation objects"
            )

        self.set_list_attribute(other, Annotation, "annotations")
        self.set_single_attribute(other, Signature, "signature")
        if isinstance(other, ItemData):
            if other.itemoid in self.items:
                raise ValueError(
                    "ItemGroupData object with that itemoid is already in the ItemGroupData object"
                )
            self.items[other.itemoid] = other
        return other


class ItemData(TransactionalElement, LastUpdateMixin, MilestoneMixin):
    """Models the ODM ItemData object"""

    ALLOWED_TRANSACTION_TYPES = ["Insert", "Update", "Upsert", "Context", "Remove"]

    def __init__(
        self,
        itemoid,
        value=Unset,
        specify_value=None,
        transaction_type=None,
        lock=None,
        freeze=None,
        verify=None,
    ):
        """
        :param str itemoid: OID for the matching :class:`ItemDef`
        :param str value: Value for the ItemData
        :param str specify_value: 'If other, specify' value - *Rave specific attribute*
        :param str transaction_type: Transaction Type for Data (one of **Insert**, **Update**, **Upsert**, **Context**, **Remove**)
        :param bool lock: Lock the DataPoint? - *Rave specific attribute*
        :param bool freeze: Freeze the DataPoint? - *Rave specific attribute*
        :param bool verify: Verify the DataPoint? - *Rave specific attribute*
        """
        super(self.__class__, self).__init__(transaction_type)
        self.itemoid = itemoid
        self.value = value

        self.specify_value = specify_value
        self.lock = lock
        self.freeze = freeze
        self.verify = verify
        #: the corresponding :class:`AuditRecord` for the DataPoint
        self.audit_record = None
        #: the list of :class:`MdsolQuery` on the DataPoint  - *Rave Specific Attribute*
        self.queries = []
        #: the list of :class:`Annotation` on the DataPoint - *Not supported by Rave*
        self.annotations = []
        #: the corresponding :class:`MeasurementUnitRef` for the DataPoint
        self.measurement_unit_ref = None
        #: the list of :class:`MdsolProtocolDeviation` references on the DataPoint - *Rave Specific Attribute*
        self.deviations = []

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        params = dict(ItemOID=self.itemoid)

        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        if self.value is not Unset:
            if self.value in [None, ""]:
                params["IsNull"] = "Yes"
            else:
                params["Value"] = str(self.value)

        if self.specify_value is not None:
            params["mdsol:SpecifyValue"] = self.specify_value

        if self.lock is not None:
            params["mdsol:Lock"] = bool_to_yes_no(self.lock)

        if self.freeze is not None:
            params["mdsol:Freeze"] = bool_to_yes_no(self.freeze)

        if self.verify is not None:
            params["mdsol:Verify"] = bool_to_yes_no(self.verify)

        # mixins
        self.mixin()
        self.mixin_params(params)

        builder.start("ItemData", params)

        if self.audit_record is not None:
            self.audit_record.build(builder)

        # Measurement unit ref must be after audit record or RWS complains
        if self.measurement_unit_ref is not None:
            self.measurement_unit_ref.build(builder)

        for query in self.queries:  # type: MdsolQuery
            query.build(builder)

        for deviation in self.deviations:  # type: MdsolProtocolDeviation
            deviation.build(builder)

        for annotation in self.annotations:  # type: Annotation
            annotation.build(builder)

        builder.end("ItemData")

    def __lshift__(self, other):
        if not isinstance(
            other,
            (
                MeasurementUnitRef,
                AuditRecord,
                MdsolQuery,
                Annotation,
                MdsolProtocolDeviation,
            ),
        ):
            raise ValueError(
                "ItemData object can only receive MeasurementUnitRef, AuditRecord, Annotation,"
                "MdsolProtocolDeviation or MdsolQuery objects"
            )
        self.set_single_attribute(other, MeasurementUnitRef, "measurement_unit_ref")
        self.set_single_attribute(other, AuditRecord, "audit_record")
        self.set_list_attribute(other, MdsolQuery, "queries")
        self.set_list_attribute(other, MdsolProtocolDeviation, "deviations")
        self.set_list_attribute(other, Annotation, "annotations")
        return other


class Signature(ODMElement):
    """
    An electronic signature applies to a collection of clinical data.
    This indicates that some user accepts legal responsibility for that data.
    See 21 CFR Part 11.
    The signature identifies the person signing, the location of signing,
    the signature meaning (via the referenced SignatureDef),
    the date and time of signing,
    and (in the case of a digital signature) an encrypted hash of the included data.
    """

    def __init__(
        self,
        signature_id=None,
        user_ref=None,
        location_ref=None,
        signature_ref=None,
        date_time_stamp=None,
    ):
        #: Unique ID for Signature
        """
        :param UserRef user_ref: :class:`UserRef` for :class:`User` signing Data
        :param LocationRef location_ref: :class:`LocationRef` for :class:`Location` of signing
        :param SignatureRef signature_ref: :class:`SignatureRef` for :class:`SignatureDef` providing signature meaning
        :param date_time_stamp: :class:`DateTimeStamp` for the time of Signature
        """
        self._id = None
        if signature_id:
            self.signature_id = signature_id
        self.user_ref = user_ref
        self.location_ref = location_ref
        self.signature_ref = signature_ref
        self.date_time_stamp = date_time_stamp

    @property
    def signature_id(self):
        """
        The ID for the Signature

        .. note:: If a Signature element is contained within a Signatures element, the ID attribute is required.
        """
        return self._id

    @signature_id.setter
    def signature_id(self, id):
        """Set the ID for the Signature"""
        self._id = id

    def build(self, builder):
        """
        Build XML by appending to builder
        """

        params = {}
        if self.signature_id is not None:
            # If a Signature element is contained within a Signatures element, the ID attribute is required.
            params["ID"] = self.signature_id

        builder.start("Signature", params)

        if self.user_ref is None:
            raise ValueError("User Reference not set.")
        self.user_ref.build(builder)

        if self.location_ref is None:
            raise ValueError("Location Reference not set.")
        self.location_ref.build(builder)

        if self.signature_ref is None:
            raise ValueError("Signature Reference not set.")
        self.signature_ref.build(builder)

        if self.date_time_stamp is None:
            raise ValueError("DateTime not set.")
        self.date_time_stamp.build(builder)

        builder.end("Signature")

    def __lshift__(self, other):
        if not isinstance(other, (UserRef, LocationRef, SignatureRef, DateTimeStamp)):
            raise ValueError(
                "Signature cannot accept a child element of type %s"
                % other.__class__.__name__
            )

        # Order is important, apparently
        self.set_single_attribute(other, UserRef, "user_ref")
        self.set_single_attribute(other, LocationRef, "location_ref")
        self.set_single_attribute(other, SignatureRef, "signature_ref")
        self.set_single_attribute(other, DateTimeStamp, "date_time_stamp")
        return other


class Annotation(TransactionalElement):
    """
    A general note about clinical data.
    If an annotation has both a comment and flags, the flags should be related to the comment.

    .. note:: Annotation is not supported by Medidata Rave
    """

    ALLOWED_TRANSACTION_TYPES = ["Insert", "Update", "Remove", "Upsert", "Context"]

    def __init__(
        self,
        annotation_id=None,
        seqnum=1,
        flags=None,
        comment=None,
        transaction_type=None,
    ):
        """
        :param id: ID for this Annotation (required if contained within an Annotations element)
        :type id: str or None
        :param int seqnum: :attr:`SeqNum` for Annotation
        :param flags: one or more :class:`Flag` for the Annotation
        :type flags: Flag or list(Flag)
        :param comment: one or more :class:`Comment` for the Annotation
        :type comment: Comment
        :param transaction_type: :attr:`TransactionType` for Annotation (one of **Insert**, **Update**, *Remove*, **Upsert**, **Context**)
        """
        super(Annotation, self).__init__(transaction_type=transaction_type)
        # initialise the flags collection
        self.flags = []
        if flags:
            if isinstance(flags, (list, tuple)):
                for flag in flags:
                    self << flag
            elif isinstance(flags, Flag):
                self << flags
            else:
                raise AttributeError("Flags attribute should be an iterable or Flag")
        self._id = None
        if annotation_id is not None:
            self.annotation_id = annotation_id
        self._seqnum = None
        if seqnum is not None:
            # validate the input
            self.seqnum = seqnum
        self.comment = comment

    @property
    def annotation_id(self):
        """
        ID for annotation

        .. note:: If an Annotation is contained with an Annotations element, the ID attribute is required.
        """
        return self._id

    @annotation_id.setter
    def annotation_id(self, value):
        """Set ID for Annotation"""
        if value in [None, ""] or str(value).strip() == "":
            raise AttributeError("Invalid ID value supplied")
        self._id = value

    @property
    def seqnum(self):
        """
        SeqNum attribute (a small positive integer) uniquely identifies the annotation within its parent entity.
        """
        return self._seqnum

    @seqnum.setter
    def seqnum(self, value):
        """
        Set SeqNum for Annotation
        :param value: SeqNum value
        :type value: int
        """
        if not re.match(r"\d+", str(value)) or value < 0:
            raise AttributeError("Invalid SeqNum value supplied")
        self._seqnum = value

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        params = {}

        # Add in the transaction type
        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        if self.seqnum is None:
            # SeqNum is not optional (and defaulted)
            raise ValueError("SeqNum is not set.")  # pragma: no cover
        params["SeqNum"] = str(self.seqnum)

        if self.annotation_id is not None:
            # If an Annotation is contained with an Annotations element,
            # the ID attribute is required.
            params["ID"] = self.annotation_id

        builder.start("Annotation", params)

        if self.flags in (None, []):
            raise ValueError("Flag is not set.")

        # populate the flags
        for flag in self.flags:
            flag.build(builder)

        # add the Comment, if it exists
        if self.comment is not None:
            self.comment.build(builder)

        builder.end("Annotation")

    def __lshift__(self, other):
        if not isinstance(other, (Flag, Comment)):
            raise ValueError(
                "Annotation cannot accept a child element of type %s"
                % other.__class__.__name__
            )

        self.set_single_attribute(other, Comment, "comment")
        self.set_list_attribute(other, Flag, "flags")
        return other


class Annotations(ODMElement):
    """
    Groups Annotation elements referenced by ItemData[TYPE] elements.
    """

    def __init__(self, annotations=[]):
        self.annotations = []
        for annotation in annotations:
            self << annotation

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        builder.start("Annotations", {})

        # populate the flags
        for annotation in self.annotations:
            annotation.build(builder)

        builder.end("Annotations")

    def __lshift__(self, other):
        if not isinstance(other, (Annotation,)):
            raise ValueError(
                "Annotations cannot accept a child element of type %s"
                % other.__class__.__name__
            )

        self.set_list_attribute(other, Annotation, "annotations")
        return other


class Comment(ODMElement):
    """
    A free-text (uninterpreted) comment about clinical data.
    The comment may have come from the Sponsor or the clinical Site.

    .. note:: Comment is not supported by Medidata Rave
    """

    VALID_SPONSOR_OR_SITE_RESPONSES = ["Sponsor", "Site"]

    def __init__(self, text=None, sponsor_or_site=None):
        """
        :param str text: Text for Comment
        :param str sponsor_or_site: Originator flag for Comment (either _Sponsor_ or _Site_)
        """
        self._text = text
        self._sponsor_or_site = sponsor_or_site

    @property
    def text(self):
        """Text content of Comment"""
        return self._text

    @text.setter
    def text(self, value):
        """Set Text content for Comment (validation of input)"""
        if value in (None, "") or value.strip() == "":
            raise AttributeError("Empty text value is invalid.")
        self._text = value

    @property
    def sponsor_or_site(self):
        """Originator of comment (either Sponsor or Site)"""
        return self._sponsor_or_site

    @sponsor_or_site.setter
    def sponsor_or_site(self, value):
        """Set Originator with validation of input"""
        if value not in Comment.VALID_SPONSOR_OR_SITE_RESPONSES:
            raise AttributeError(
                "%s sponsor_or_site value of %s is not valid"
                % (self.__class__.__name__, value)
            )
        self._sponsor_or_site = value

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        if self.text is None:
            raise ValueError("Text is not set.")
        params = {}
        if self.sponsor_or_site is not None:
            params["SponsorOrSite"] = self.sponsor_or_site

        builder.start("Comment", params)
        builder.data(self.text)
        builder.end("Comment")


class Flag(ODMElement):
    """
    A machine-processable annotation on clinical data.

    .. note:: Flag is not supported by Rave
    """

    def __init__(self, flag_type=None, flag_value=None):
        """
        :param FlagType flag_type: Type for Flag
        :param FlagValue flag_value: Value for Flag
        """
        self.flag_type = None
        self.flag_value = None
        if flag_type is not None:
            self << flag_type
        if flag_value is not None:
            self << flag_value

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        builder.start("Flag", {})

        if self.flag_type is not None:
            self.flag_type.build(builder)

        if self.flag_value is None:
            raise ValueError("FlagValue is not set.")
        self.flag_value.build(builder)

        builder.end("Flag")

    def __lshift__(self, other):
        if not isinstance(other, (FlagType, FlagValue)):
            raise ValueError(
                "Flag cannot accept a child element of type %s"
                % other.__class__.__name__
            )

        # Order is important, apparently
        self.set_single_attribute(other, FlagType, "flag_type")
        self.set_single_attribute(other, FlagValue, "flag_value")
        return other


class FlagType(ODMElement):
    """
    The type of flag. This determines the purpose and semantics of the flag.
    Different applications are expected to be interested in different types of flags.
    The actual value must be a member of the referenced CodeList.

    .. note:: FlagType is not supported by Rave
    """

    def __init__(self, flag_type, codelist_oid=None):
        """
        :param flag_type: Type for :class:`Flag`
        """
        self.flag_type = flag_type
        self._codelist_oid = None
        if codelist_oid is not None:
            self.codelist_oid = codelist_oid

    @property
    def codelist_oid(self):
        """Reference to the :class:`CodeList` for the FlagType"""
        return self._codelist_oid

    @codelist_oid.setter
    def codelist_oid(self, value):
        if value in (None, "") or value.strip() == "":
            raise AttributeError("Empty CodeListOID value is invalid.")
        self._codelist_oid = value

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        if self.codelist_oid is None:
            raise ValueError("CodeListOID not set.")
        builder.start("FlagType", dict(CodeListOID=self.codelist_oid))
        builder.data(self.flag_type)
        builder.end("FlagType")


class FlagValue(ODMElement):
    """
    The value of the flag. The meaning of this value is typically dependent on the associated FlagType.
    The actual value must be a member of the referenced CodeList.

    .. note::  FlagValue is not supported by Rave
    """

    def __init__(self, flag_value, codelist_oid=None):
        """
        :param flag_value: Value for :class:`Flag`
        """
        self.flag_value = flag_value
        self._codelist_oid = None
        if codelist_oid is not None:
            self.codelist_oid = codelist_oid

    @property
    def codelist_oid(self):
        """Reference to the :class:`CodeList` for the FlagType"""
        return self._codelist_oid

    @codelist_oid.setter
    def codelist_oid(self, value):
        if value in (None, "") or value.strip() == "":
            raise AttributeError("Empty CodeListOID value is invalid.")
        self._codelist_oid = value

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        if self.codelist_oid is None:
            raise ValueError("CodeListOID not set.")
        builder.start("FlagValue", dict(CodeListOID=self.codelist_oid))
        builder.data(self.flag_value)
        builder.end("FlagValue")


class UserRef(ODMElement):
    """
    Reference to a :class:`User`
    """

    def __init__(self, oid):
        """
        :param str oid: OID for referenced :class:`User`
        """
        self.oid = oid

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        builder.start("UserRef", dict(UserOID=self.oid))
        builder.end("UserRef")


class LocationRef(ODMElement):
    """
    Reference to a :class:`Location`
    """

    def __init__(self, oid):
        """
        :param str oid: OID for referenced :class:`Location`
        """
        self.oid = oid

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        builder.start("LocationRef", dict(LocationOID=str(self.oid)))
        builder.end("LocationRef")


class SiteRef(ODMElement, LastUpdateMixin):
    """
    Reference to a :class:`Location`
    The default value is `SiteName`, and the value `SiteUUID` implies that the `LocationOID`
    .. note:: The `mdsol:LocationOIDType` attribute should be used to indicate the type of `LocationOID`
    """

    def __init__(self,
                 oid,
                 location_oid_type: Optional[LocationOIDType] = None,
                 site_number: Optional[str] = None,
                 study_site_name: Optional[str] = None,
                 site_uuid: Optional[str] = None,
                 location_name: Optional[str] = None,
                 study_env_site_number: Optional[str] = None,
                 previous_location_oid: Optional[str] = None,
                 shared_location_oid: Optional[str] = None,
                 previous_shared_location_oid: Optional[str] = None,
                 subject_created_at_site: Optional[str] = None,):
        """
        :param str oid: OID for referenced :class:`Location`
        :param LocationOIDType location_oid_type: Type for the oid
        :param site_number: Site number
        :param study_site_name: Study site name
        :param site_uuid: Site UUID
        :param location_name: Location name
        :param study_env_site_number: Study environment site number
        :param previous_location_oid: Previous location OID
        :param shared_location_oid: Shared location OID
        :param previous_shared_location_oid: Previous shared location OID
        :param subject_created_at_site: Subject created at site
        """
        self.oid = oid
        self.location_oid_type = location_oid_type
        self.site_number = site_number
        self.study_site_name = study_site_name
        self.site_uuid = site_uuid
        self.location_name = location_name
        self.study_env_site_number = study_env_site_number
        self.previous_location_oid = previous_location_oid
        self.shared_location_oid = shared_location_oid
        self.previous_shared_location_oid = previous_shared_location_oid
        self.subject_created_at_site = subject_created_at_site

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        params = dict(LocationOID=self.oid)
        if self.location_oid_type:
            params["mdsol:LocationOIDType"] = self.location_oid_type.value
        elif self.site_uuid:
            params["mdsol:SiteUUID"] = self.site_uuid
        elif self.site_number:
            params["mdsol:SiteNumber"] = self.site_number
        elif self.study_site_name:
            params["mdsol:StudySiteName"] = self.study_site_name
        elif self.study_env_site_number:
            params["mdsol:StudyEnvSiteNumber"] = self.study_env_site_number
        if self.location_name:
            params["mdsol:LocationName"] = self.location_name
        if self.previous_location_oid:
            params["mdsol:PreviousLocationOID"] = self.previous_location_oid
        if self.shared_location_oid:
            params["mdsol:SharedLocationOID"] = self.shared_location_oid
        if self.subject_created_at_site:
            params["mdsol:SubjectCreatedAtSite"] = self.subject_created_at_site
        # mixins
        self.mixin()
        self.mixin_params(params)

        builder.start("SiteRef", params)
        builder.end("SiteRef")


class SignatureRef(ODMElement):
    """
    Reference to a Signature
    """

    def __init__(self, oid):
        """
        :param str oid: OID for referenced :class:`Signature`
        """
        self.oid = oid

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        builder.start("SignatureRef", dict(SignatureOID=self.oid))
        builder.end("SignatureRef")


class ReasonForChange(ODMElement):
    """
    A user-supplied reason for a data change.
    """

    def __init__(self, reason):
        """
        :param str reason: Supplied Reason for change
        """
        self.reason = reason

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        builder.start("ReasonForChange", {})
        builder.data(self.reason)
        builder.end("ReasonForChange")


class DateTimeStamp(ODMElement):
    """
    The date/time that the data entry, modification, or signature was performed.
    This applies to the initial occurrence of the action, not to subsequent transfers between computer systems.
    """

    def __init__(self, date_time):
        #: specified DateTime for event
        self.date_time = date_time

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        builder.start("DateTimeStamp", {})
        if isinstance(self.date_time, datetime):
            builder.data(dt_to_iso8601(self.date_time))
        else:
            builder.data(self.date_time)
        builder.end("DateTimeStamp")


class SourceID(ODMElement):
    """
    Information that identifies the source of the data within an originating system. 
      It is only meaningful within the context of that system.
    """

    def __init__(self, source_id):
        #: specified DateTime for event
        self.source_id = source_id

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        builder.start("SourceID", {})
        builder.data(self.source_id)
        builder.end("SourceID")


class AuditRecord(ODMElement):
    """
    An AuditRecord carries information pertaining to the creation, deletion, or modification of clinical data.
    This information includes who performed that action, and where, when, and why that action was performed.

    .. note:: AuditRecord is supported only by :class:`ItemData` in Rave
    """

    EDIT_MONITORING = "Monitoring"
    EDIT_DATA_MANAGEMENT = "DataManagement"
    EDIT_DB_AUDIT = "DBAudit"
    EDIT_POINTS = [EDIT_MONITORING, EDIT_DATA_MANAGEMENT, EDIT_DB_AUDIT]

    def __init__(
        self,
        edit_point=None,
        used_imputation_method=None,
        identifier=None,
        include_file_oid=None,
    ):
        """
        :param str identifier: Audit identifier
        :param str edit_point: EditPoint attribute identifies the phase of data processing in which action occurred
            (*Monitoring*, *DataManagement*, *DBAudit*)
        :param bool used_imputation_method: Indicates whether the action involved the use of a Method
        :param bool include_file_oid: Include the FileOID in the AuditRecord
        """
        self._edit_point = None
        self.edit_point = edit_point
        self.used_imputation_method = used_imputation_method
        self._id = None
        if identifier:
            self.audit_id = identifier
        self.include_file_oid = include_file_oid
        #: :class:`UserRef` for the AuditRecord
        self.user_ref = None
        #: :class:`LocationRef` for the AuditRecord
        self.location_ref = None
        #: :class:`ReasonForChange` for the AuditRecord
        self.reason_for_change = None
        #: :class:`DateTimeStamp` for the AuditRecord
        self.date_time_stamp = None
        #: :class:`SourceID` for the AuditRecord
        self.source_id = None

    @property
    def audit_id(self):
        """
        AuditRecord ID

        .. note:: If an AuditRecord is contained within an AuditRecords element, the ID attribute must be provided.
        """
        return self._id

    @audit_id.setter
    def audit_id(self, value):
        if value not in [None, ""] and str(value).strip() != "":
            val = str(value).strip()[0]
            if val not in VALID_ID_CHARS:
                raise AttributeError(
                    '%s id cannot start with "%s" character'
                    % (self.__class__.__name__, val)
                )
        self._id = value

    @property
    def edit_point(self):
        """
        EditPoint attribute identifies the phase of data processing in which action occurred
            (*Monitoring*, *DataManagement*, *DBAudit*)
        """
        return self._edit_point

    @edit_point.setter
    def edit_point(self, value):
        if value is not None:
            if value not in self.EDIT_POINTS:
                raise AttributeError(
                    "%s edit_point must be one of %s not %s"
                    % (self.__class__.__name__, ",".join(self.EDIT_POINTS), value)
                )
        self._edit_point = value

    def build(self, builder):
        """Build XML by appending to builder"""
        params = {}

        if self.edit_point is not None:
            params["EditPoint"] = self.edit_point

        if self.used_imputation_method is not None:
            params["UsedImputationMethod"] = bool_to_yes_no(self.used_imputation_method)

        if self.audit_id is not None:
            params["ID"] = str(self.audit_id)

        if self.include_file_oid is not None:
            params["mdsol:IncludeFileOID"] = bool_to_yes_no(self.include_file_oid)

        builder.start("AuditRecord", params)
        if self.user_ref is None:
            raise ValueError("User Reference not set.")
        self.user_ref.build(builder)

        if self.location_ref is None:
            raise ValueError("Location Reference not set.")
        self.location_ref.build(builder)

        if self.date_time_stamp is None:
            raise ValueError("DateTime not set.")

        self.date_time_stamp.build(builder)

        # Optional
        if self.source_id:
            self.source_id.build(builder)
        # Optional
        if self.reason_for_change is not None:
            self.reason_for_change.build(builder)

        builder.end("AuditRecord")

    def __lshift__(self, other):
        if not isinstance(
            other, (UserRef, LocationRef, DateTimeStamp, ReasonForChange, SourceID)
        ):
            raise ValueError(
                "AuditRecord cannot accept a child element of type %s"
                % other.__class__.__name__
            )

        # Order is important, apparently
        self.set_single_attribute(other, UserRef, "user_ref")
        self.set_single_attribute(other, LocationRef, "location_ref")
        self.set_single_attribute(other, DateTimeStamp, "date_time_stamp")
        self.set_single_attribute(other, ReasonForChange, "reason_for_change")
        self.set_single_attribute(other, SourceID, "source_id")
        return other


class MdsolProtocolDeviation(TransactionalElement):
    """
    Extension for Protocol Deviations in Rave

    .. note::
        * This is a Medidata Rave Specific Extension
        * This primarily exists as a mechanism for use by the Clinical Audit Record Service, but it is useful to define for the builders

    """

    ALLOWED_TRANSACTION_TYPES = ["Insert"]

    def __init__(
        self, value, status, repeat_key=1, code=None, klass=None, transaction_type=None
    ):
        """
        :param str value: Value for the Protocol Deviation 
        :param rwslib.builder_constants.ProtocolDeviationStatus status: 
        :param int repeat_key: RepeatKey for the Protocol Deviation
        :param basestring code: Protocol Deviation Code
        :param basestring klass: Protocol Deviation Class 
        :param transaction_type: Transaction Type for the Protocol Deviation
        """
        super(MdsolProtocolDeviation, self).__init__(transaction_type=transaction_type)
        self._status = None
        self._repeat_key = None
        self.status = status
        self.value = value
        self.repeat_key = repeat_key
        self.code = code
        self.pdclass = klass

    @property
    def repeat_key(self):
        return self._repeat_key

    @repeat_key.setter
    def repeat_key(self, value):
        if isinstance(value, int):
            self._repeat_key = value
        else:
            raise ValueError("RepeatKey should be an integer, not {}".format(value))

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        if isinstance(value, ProtocolDeviationStatus):
            self._status = value
        else:
            raise ValueError(
                "Status {} is not a valid ProtocolDeviationStatus".format(value)
            )

    def build(self, builder):
        """Build XML by appending to builder"""
        params = dict(
            Value=self.value,
            Status=self.status.value,
            ProtocolDeviationRepeatKey=self.repeat_key,
        )

        if self.code:
            params["Code"] = self.code
        if self.pdclass:
            params["Class"] = self.pdclass
        if self.transaction_type:
            params["TransactionType"] = self.transaction_type
        builder.start("mdsol:ProtocolDeviation", params)
        builder.end("mdsol:ProtocolDeviation")


class MdsolQuery(ODMElement):
    """
    MdsolQuery extension element for Queries at item level only

    .. note:: This is a Medidata Rave specific extension
    """

    def __init__(
        self,
        value=None,
        query_repeat_key=None,
        recipient=None,
        status=None,
        requires_response=None,
        response=None,
        preceding_query_repeat_key=None,
        ):
        """
        :param str value: Query Value
        :param int query_repeat_key: Repeat key for Query
        :param str recipient: Recipient for Query (Marking Group)
        :param QueryStatusType status: Query status
        :param bool requires_response: Does this Query need a response?
        :param response: Query response (if any)
        :type response: str or None
        :param int preceding_query_repeat_key: Query Key returned by OpenQuery call
        """
        self.value = value
        self.query_repeat_key = query_repeat_key
        self.recipient = recipient
        self._status = None
        self.status = status
        self.requires_response = requires_response
        self.response = response
        self.preceding_query_repeat_key = preceding_query_repeat_key

    @property
    def status(self):
        """Query Status"""
        return self._status

    @status.setter
    def status(self, value):
        """Set Query Status"""
        if value is not None:
            if not isinstance(value, QueryStatusType):
                raise AttributeError(
                    "%s action type is invalid in mdsol:Query." % (value,)
                )
        self._status = value

    def build(self, builder):
        """
        Build XML by appending to builder
        """
        params = {}

        if self.value is not None:
            params["Value"] = str(self.value)

        if self.query_repeat_key is not None:
            params["QueryRepeatKey"] = str(self.query_repeat_key)

        if self.recipient is not None:
            params["Recipient"] = str(self.recipient)

        if self.status is not None:
            params["Status"] = self.status.value

        if self.requires_response is not None:
            params["RequiresResponse"] = bool_to_yes_no(self.requires_response)

        # When closing a query
        if self.response is not None:
            params["Response"] = str(self.response)

        # When re-opening a query
        if self.preceding_query_repeat_key is not None:
            params["PrecedingQueryRepeatKey"] = str(self.preceding_query_repeat_key)

        builder.start("mdsol:Query", params)
        builder.end("mdsol:Query")



# -*- coding: utf-8 -*-
import platform
from string import ascii_letters
import datetime
from xml.etree import cElementTree as ET


# -----------------------------------------------------------------------------------------------------------------------
# Constants

VALID_ID_CHARS = ascii_letters + '_'


# -----------------------------------------------------------------------------------------------------------------------
# Utilities
def get_utc_date() -> datetime.datetime:
    """
    Returns the UTC date as datetime.datetime object.
    """
    version = platform.python_version_tuple()
    if int(version[1]) < 11:
        utc_date = datetime.datetime.utcnow()
    else:
        utc_date = datetime.datetime.now(datetime.UTC)
    return utc_date


def now_to_iso8601():
    """Returns NOW date/time as a UTC date/time formated as iso8601 string"""
    utc_date = get_utc_date()
    return dt_to_iso8601(utc_date)


def dt_to_iso8601(dt):
    """Turn a datetime into an ISO8601 formatted string"""
    return dt.strftime("%Y-%m-%dT%H:%M:%S")


def bool_to_yes_no(val):
    """Convert True/False to Yes/No"""
    return 'Yes' if val else 'No'


def bool_to_true_false(val):
    """Convert True/False to TRUE / FALSE"""
    return 'TRUE' if val else 'FALSE'


def indent(elem, level=0):
    """Indent a elementree structure"""
    i = "\n" + level * "  "
    if len(elem) > 0:
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def make_element(builder, tag, content):
    """Make an element with this tag and text content"""
    builder.start(tag, {})
    builder.data(content)  # Must be UTF-8 encoded
    builder.end(tag)


# -----------------------------------------------------------------------------------------------------------------------
# Classes

class ODMElement(object):
    """Base class for ODM XML element classes"""

    def __call__(self, *args):
        """Collect all children passed in call"""
        for child in args:
            self << child
        return self

    def __lshift__(self, other):
        """__lshift__ should be overridden in descendant classes to accept child elements and incorporate them.
           By default takes no child elements
        """
        raise ValueError("%s takes no child elements" % self.__class__.__name__)

    def add(self, *args):
        """Like call but adds a set of args"""
        for child in args:
            self << child
        return self

    def __str__(self):
        """Return string representation"""
        builder = ET.TreeBuilder()
        self.build(builder)
        return ET.tostring(builder.close(), encoding='utf-8').decode('utf-8')

    def set_single_attribute(self, other, trigger_klass, property_name):
        """Used to set guard the setting of an attribute which is singular and can't be set twice"""

        if isinstance(other, trigger_klass):

            # Check property exists
            if not hasattr(self, property_name):
                raise AttributeError("%s has no property %s" % (self.__class__.__name__, property_name))

            if getattr(self, property_name) is None:
                setattr(self, property_name, other)
            else:
                raise ValueError(
                    '%s already has a %s element set.' % (self.__class__.__name__, other.__class__.__name__,))

    def set_list_attribute(self, other, trigger_klass, property_name):
        """Used to set guard the setting of a list attribute, ensuring the same element is not added twice."""
        # Check property exists
        if isinstance(other, trigger_klass):

            if not hasattr(self, property_name):
                raise AttributeError("%s has no property %s" % (self.__class__.__name__, property_name))

            val = getattr(self, property_name, [])
            if other in val:
                raise ValueError("%s already exists in %s" % (other.__class__.__name__, self.__class__.__name__))
            else:
                val.append(other)
                setattr(self, property_name, val)


class TransactionalElement(ODMElement):
    """
    Models an ODM Element that is allowed a transaction type. Different elements have different
    allowed transaction types
    """
    ALLOWED_TRANSACTION_TYPES = []

    def __init__(self, transaction_type):
        self._transaction_type = None
        self.transaction_type = transaction_type

    @property
    def transaction_type(self):
        """returns the TransactionType attribute"""
        return self._transaction_type

    @transaction_type.setter
    def transaction_type(self, value):
        """Set the TransactionType (with Input Validation)"""
        if value is not None:
            if value not in self.ALLOWED_TRANSACTION_TYPES:
                raise AttributeError('%s transaction_type element must be one of %s not %s' % (
                    self.__class__.__name__, ','.join(self.ALLOWED_TRANSACTION_TYPES), value,))
        self._transaction_type = value


class Unset:
    """
    A class to represent an unset value.
    """
    def __bool__(self) -> bool:
        return False



# -*- coding: utf-8 -*-

__author__ = 'glow'

# Common elements
from rwslib.builders.common import ODMElement, TransactionalElement

# Constants
from rwslib.builders.constants import *

# Core
from rwslib.builders.core import *

# Metadata
from rwslib.builders.metadata import *

# ClinicalData
from rwslib.builders.clinicaldata import *

# AdminData
from rwslib.builders.admindata import *

# ReferenceData
#from rwslib.builders.referencedata import *

# MODM
from rwslib.builders.modm import *





# -*- coding: utf-8 -*-
__author__ = 'isparks'

"""Holds all context classes used by the parser"""


class ContextBase(object):
    """Base class"""

    def __repr__(self):
        vals = dict((k, v) for k, v in self.__dict__.items() if v is not None)
        return "{0}({1})".format(self.__class__.__name__, str(vals))


class Context(ContextBase):
    """A context class that holds the data for every reportable element in the ODM response"""

    def __init__(self, study_oid, subcategory, metadata_version):
        self.study_oid = study_oid
        self.subcategory = subcategory
        self.metadata_version = metadata_version
        self.subject = None
        self.event = None
        self.form = None
        self.itemgroup = None
        self.item = None
        self.audit_record = AuditRecord()
        self.query = None
        self.protocol_deviation = None
        self.comment = None
        self.review = None
        self.signature = Signature()


class AuditRecord(ContextBase):
    def __init__(self):
        self.user_oid = None
        self.location_oid = None
        self.datetimestamp = None
        self.reason_for_change = None
        self.source_id = -1


class Signature(ContextBase):
    def __init__(self):
        self.oid = None
        self.user_oid = None
        self.location_oid = None
        self.datetimestamp = None


class Subject(ContextBase):
    def __init__(self, key, name, status, transaction_type):
        self.key = key
        self.name = name
        self.status = status
        self.transaction_type = transaction_type


class ContextContainer(ContextBase):
    """Base classes for ODM containers that have an oid and repeat key"""

    def __init__(self, oid, repeat_key, transaction_type):
        self.oid = oid
        self.repeat_key = repeat_key
        self.transaction_type = transaction_type


class Event(ContextContainer):
    def __init__(self, oid, repeat_key, transaction_type, instance_name, instance_overdue, instance_id):
        ContextContainer.__init__(self, oid, repeat_key, transaction_type)
        self.instance_name = instance_name
        self.instance_overdue = instance_overdue
        self.instance_id = instance_id


class Form(ContextContainer):
    def __init__(self, oid, repeat_key, transaction_type, datapage_name, datapage_id):
        ContextContainer.__init__(self, oid, repeat_key, transaction_type)
        self.datapage_name = datapage_name
        self.datapage_id = datapage_id


class ItemGroup(ContextContainer):
    def __init__(self, oid, repeat_key, transaction_type, record_id):
        ContextContainer.__init__(self, oid, repeat_key, transaction_type)
        self.record_id = record_id


class Item(ContextBase):
    def __init__(self, oid, value, specify_value, signature_broken, freeze, verify, lock, transaction_type):
        self.oid = oid
        self.value = value
        self.specify_value = specify_value
        self.signature_broken = signature_broken
        self.freeze = freeze
        self.verify = verify
        self.lock = lock
        self.transaction_type = transaction_type


class Query(ContextBase):
    """Query related attributes"""

    def __init__(self, repeat_key, status, response, recipient, value):
        self.repeat_key = repeat_key
        self.status = status
        self.response = response
        self.recipient = recipient
        self.value = value


class Review(ContextBase):
    """Review related attributes"""

    def __init__(self, group_name, reviewed):
        self.group_name = group_name
        self.reviewed = reviewed


class Comment(ContextBase):
    """Review related attributes"""

    def __init__(self, repeat_key, value, transaction_type):
        self.repeat_key = int(repeat_key) if repeat_key else repeat_key
        self.value = value
        self.transaction_type = transaction_type


class ProtocolDeviation(ContextBase):
    """Protocol Deviation related attributes"""

    def __init__(self, repeat_key, code, klass, status, value, transaction_type):
        self.repeat_key = repeat_key
        self.code = code
        self.klass = klass
        self.status = status
        self.value = value
        self.transaction_type = transaction_type



__author__ = 'isparks'

# -----------------------------------------------------------------------------------------------------------------------
# Overview
#
# local_cv is a small framework utilizing rwslib that pulls clinical view data from Rave and into a local
# (to your own machine) database where it can be queried, reported on by SQL tools etc.
#
# The library provides a simple SQLLite proof of concept but other SQL engines could be supported by creating
# new subclasses of BaseDBAdapter to handle the DDL, DML and database population.

from rwslib import RWSConnection
from rwslib.rws_requests.biostats_gateway import ProjectMetaDataRequest, FormDataRequest

import os
import csv
from itertools import groupby
import sqlite3
import logging
import re

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

VIEW_REX = re.compile(r"V_(?P<study>[A-Za-z0-9]+)_(?P<view>[A-Za-z0-9]+)(?:_(?P<type>[A-Za-z0-9]+))?")


# -----------------------------------------------------------------------------------------------------------------------
# Classes

class BaseDBAdapter(object):
    """
    A DBAdapter provides functionality to convert a set of dataset definitions into DDL statements suitable for
    storage of data in that format. It also provides capability to insert data into that database once provided.

    Key responsibilities:

    1. Convert set of dataset definitions to DDL
    2. Populate / create tables with DDL
    3. Convert set of data rows to INSERT statements
    4. Execute Insert statements against database to build DB

    """

    def __init__(self):
        """Override in descendant classes to receive information about the DB to be populated"""
        self.datasets = {}

    @staticmethod
    def getCSVReader(data, reader_type=csv.DictReader):
        """Take a Rave CSV output ending with a line with just EOF on it and return a DictReader"""
        f = StringIO(data[:-4])  # Remove \nEOF
        return reader_type(f)

    def processMetaData(self, metadata):
        """Takes a string representing a View metadata CSV extract from RWS, sets dataset dictionary
           where each dataset key has a list of dictionaries representing fields in dataset
        """
        self._setDatasets(metadata)
        self._processDDL()

    def _setDatasets(self, metadata):
        """Extract dataset definitions from CSV metadata"""
        # Pull the lot into a list
        cols = list(self.getCSVReader(metadata))

        # sort the cols by view name and ordinal (they come sorted, but lets be sure)
        cols.sort(key=lambda x: [x['viewname'], int(x['ordinal'])])

        # Now group them by their viewname
        for key, grp in groupby(cols, key=lambda x: x["viewname"]):
            self.datasets[key] = list(grp)

    def _processDDL(self):
        """Generate table SQL, override for each database variant"""
        raise NotImplementedError("Override _processDDL in descendant classes")

    def processFormData(self, data, dataset_name):
        """Take a string of form data as CSV and convert to insert statements, return template and data values"""

        # Get the cols for this dataset
        cols = self.datasets[dataset_name]
        reader = self.getCSVReader(data, reader_type=csv.reader)

        # Get fieldnames from first line of reader, what is left is set of rows
        fieldnames = next(reader)

        # Check columns
        for col in cols:
            varname = col["varname"]
            if varname not in fieldnames:
                raise ValueError("Column %s not found in data for dataset %s" % (varname, dataset_name,))

        # Now call overrideen methods of base classes to process this data
        self._processDML(dataset_name, cols, reader)

    def _processDML(self, dataset_name, cols, reader):
        """Create and populate table with values from dataset"""
        raise NotImplementedError("Override _processDML in descendant classes")


class SQLLiteDBAdapter(BaseDBAdapter):
    """Variant that makes SQLLite SQL"""

    def __init__(self, db):
        """Receives a SQLLiteDB to populate"""
        BaseDBAdapter.__init__(self)
        self.conn = db

    def _processDDL(self):
        """Generate and process table SQL, SQLLite version"""
        sql_statements = self._generateDDL()

        logging.info('Generating sqllite tables')
        for stmt in sql_statements:
            c = self.conn.cursor()
            c.execute(stmt)
            self.conn.commit()

    def _generateDDL(self):
        """Generate DDL statements for SQLLite"""
        sql = []
        # Next convert each set of columns into a table structure
        for dataset_name in sorted(self.datasets.keys()):

            # SQL to drop the table if it already exists
            sql.append('''drop table if exists %s;''' % dataset_name)

            # Generate the SQL for the cols
            cols = self.datasets[dataset_name]
            col_defs = []
            for col in cols:
                sql_datatype = self.getSQLDataType(col["vartype"])
                col_defs.append("%s %s" % (col["varname"], sql_datatype,))

            stmt = 'CREATE TABLE %s (%s)' % (dataset_name, ','.join(col_defs))
            sql.append(stmt)

        return sql

    @staticmethod
    def getSQLDataType(dtype):
        """Return SQLLite data type for a Rave view data type"""
        return dict(num="NUMERIC", char="TEXT")[dtype]

    def _processDML(self, dataset_name, cols, reader):
        """Overridden version of create DML for SQLLite"""
        sql_template = self._generateInsertStatement(dataset_name, cols)

        # Now insert in batch, reader is a list of rows to insert at this point
        c = self.conn.cursor()
        c.executemany(sql_template, reader)
        self.conn.commit()

    def _generateInsertStatement(self, dataset_name, cols):
        """Generates a sql INSERT template"""
        col_names = [col["varname"] for col in cols]

        # Generate question mark placeholders
        qms = ','.join(['?' for x in col_names])

        return 'INSERT INTO %s (%s) values (%s)' % (dataset_name, ','.join(col_names), qms)


class LocalCVBuilder(object):
    """
    Does the work of pulling data from Rave via RWS. All database activities are delegated to the db_adapter.
    """

    def __init__(self, rws_connection, project_name, environment, db_adapter):
        self.rws_connection = rws_connection
        self.project_name = project_name
        self.environment = environment
        self.db_adapter = db_adapter

    @staticmethod
    def name_type_from_viewname(viewname):
        """Have format V_<studyname>_<view>[_RAW], return name and view type REGULAR or RAW"""
        matched = VIEW_REX.match(viewname)
        return matched.group('view'), matched.group('type') or 'REGULAR'

    def execute(self):
        """Generate local DB, pulling metadata and data from RWSConnection"""

        logging.info('Requesting view metadata for project %s' % self.project_name)
        project_csv_meta = self.rws_connection.send_request(ProjectMetaDataRequest(self.project_name))

        # Process it into a set of tables
        self.db_adapter.processMetaData(project_csv_meta)

        # Get the data for the study
        for dataset_name in self.db_adapter.datasets.keys():
            logging.info('Requesting data from dataset %s' % dataset_name)
            form_name, _type = self.name_type_from_viewname(dataset_name)
            form_data = self.rws_connection.send_request(
                FormDataRequest(self.project_name, self.environment, _type, form_name))

            # Now process the form_data into the db of choice
            logging.info('Populating dataset %s' % dataset_name)
            self.db_adapter.processFormData(form_data, dataset_name)

        logging.info('Process complete')


if __name__ == '__main__':
    # An example
    logging.basicConfig(level=logging.INFO)

    from rwslib._settings import accounts

    # Accounts is a dict of dicts like
    # accounts = {'innovate' : {'username': 'username',
    #                          'password':'password'},
    #             'otherurl' : {'username': 'username',
    #                          'password':'password'},
    #            }

    acc = accounts['innovate']
    rv = RWSConnection('innovate', acc['username'], acc['password'])

    path, _ = os.path.split(os.path.realpath(__file__))
    db_path = os.path.join(path, 'local_cvs.db')

    # Make sqllite connection
    logging.info('Opening/overwriting sqllite db %s' % db_path)
    conn = sqlite3.connect(db_path)

    # DBAdaptor takes care of all the database specific parts of the process
    db_adaptor = SQLLiteDBAdapter(conn)

    lcv = LocalCVBuilder(rv, 'SIMPLESTUDY', 'TEST', db_adaptor)
    lcv.execute()

    # Once complete you can do further work with the conn object, run SQL statements etc.



__author__ = "isparks"

from lxml import etree

MEDI_NS = "{http://www.mdsol.com/ns/odm/metadata}"
ODM_NS = "{http://www.cdisc.org/ns/odm/v1.3}"
XLINK_NS = "{http://www.w3.org/1999/xlink}"


def getEnvironmentFromNameAndProtocol(studyname, protocolname):
    """
    Extract environment name using studyname and protocolname to guide
    :param str studyname: Name of the study (including Env)
    :param str protocolname: Name of the study
    """
    # StudyName =    "TEST (1) (DEV)"
    # ProtocolName = "TEST (1)"
    # Raw Env      =           "(DEV)"

    raw_env = studyname[len(protocolname) :].strip()

    if "(" in raw_env:
        l_brace_pos = raw_env.rfind("(")
        r_brace_pos = raw_env.rfind(")")
        return raw_env[l_brace_pos + 1 : r_brace_pos]
    else:
        return raw_env


def parseXMLString(xml):
    """
    Parse XML string, return root
    :param str: Passed in XML
    """

    # Remove BOM if it exists (different requests seem to have different BOMs)
    unichr_captured = ""

    if not xml.strip():
        return u""

    while xml[0] != u"<":
        unichr_captured += xml[0]
        xml = xml[1:]
    parser = etree.XMLParser(ns_clean=True, collect_ids=False, huge_tree=True)
    try:
        return etree.fromstring(xml.encode("utf-8"), parser=parser)
    except etree.XMLSyntaxError:
        raise Exception(xml)


class RWSException(Exception):
    """RWS Exception. Usual to attach the error response object"""

    def __init__(self, msg, rws_error):
        """
        :param str msg: Error message (base)
        :param str rws_error: RWS error message
        """
        Exception.__init__(self, msg)
        self.rws_error = rws_error


class XMLRepr(object):
    """Classes that represent objects passed back from RWS as XML"""

    def __init__(self, xml):
        """
        :param str xml: XML returned from RWS
        """
        self.root = parseXMLString(xml)

    def __unicode__(self):
        """String representation of same"""
        return etree.tostring(self.root, encoding="utf-8")

    def __str__(self):
        """String representation of same"""
        return etree.tostring(self.root, encoding="utf-8").decode("utf-8")


class ODMDoc(XMLRepr):
    """A base ODM document"""

    def __init__(self, xml):
        """
        Abstract Doc Class
        :param bytes xml: Input content
        """
        # Call base class
        XMLRepr.__init__(self, xml)
        r_get = self.root.get

        self.filetype = r_get("FileType")
        self.creationdatetime = r_get("CreationDateTime")
        self.fileoid = r_get("FileOID")
        self.ODMVersion = r_get("ODMVersion")
        self.granularity = r_get("Granularity", None)


class RWSError(ODMDoc):
    """
    Extends ODMDoc, inheriting attributes like filetype, creationdatetime etc.

    Parses XML of the form:

    .. code-block:: xml

        <?xml version="1.0" encoding="utf-8"?>
        <ODM xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata"
             FileType="Snapshot"
             CreationDateTime="2013-04-08T10:28:49.578-00:00"
             FileOID="4d13722a-ceb6-4419-a917-b6ad5d0bc30e"
             ODMVersion="1.3"
             mdsol:ErrorDescription="Incorrect login and password combination. [RWS00008]"
             xmlns="http://www.cdisc.org/ns/odm/v1.3" />
    """

    def __init__(self, xml):
        ODMDoc.__init__(self, xml)
        r_get = self.root.get
        self.errordescription = r_get(MEDI_NS + "ErrorDescription")


class RWSErrorResponse(XMLRepr):
    """
    Parses messages of the form:

    .. code-block:: xml

        <Response
            ReferenceNumber="0b47fe86-542f-4070-9e7d-16396a5ef08a"
            InboundODMFileOID="Not Supplied"
            IsTransactionSuccessful="0"
            ReasonCode="RWS00092"
            ErrorClientResponseMessage="CRF version not found">
        </Response>
    """

    def __init__(self, xml):
        # Call base class
        XMLRepr.__init__(self, xml)
        r_get = self.root.get

        self.referencenumber = r_get("ReferenceNumber")
        self.inboundodmfileoid = r_get("InboundODMFileOID")
        self.istransactionsuccessful = r_get("IsTransactionSuccessful") == "1"
        self.reasoncode = r_get("ReasonCode")
        self.errordescription = r_get("ErrorClientResponseMessage")


class RWSResponse(XMLRepr):
    """
    Parses messages of the form:

    .. code-block:: xml

        <Response ReferenceNumber="82e942b0-48e8-4cf4-b299-51e2b6a89a1b"
            InboundODMFileOID=""
            IsTransactionSuccessful="1"
            SuccessStatistics="Rave objects touched: Subjects=0; Folders=0; Forms=0; Fields=0; LogLines=0" NewRecords="">
        </Response>
    """

    def __init__(self, xml):
        # Call base class
        XMLRepr.__init__(self, xml)
        r_get = self.root.get

        self.referencenumber = r_get("ReferenceNumber")
        self.inboundodmfileoid = r_get("InboundODMFileOID")
        self.istransactionsuccessful = r_get("IsTransactionSuccessful") == "1"

        self.subjects_touched = 0
        self.folders_touched = 0
        self.forms_touched = 0
        self.fields_touched = 0
        self.loglines_touched = 0

        success_stats = r_get("SuccessStatistics", "")

        # Clinical data post
        if success_stats.startswith("Rave objects touched:"):
            success_stats = success_stats[
                len("Rave objects touched:") + 1 :
            ]  # Subjects=0; Folders=0; Forms=0; Fields=0; LogLines=0
            parts = success_stats.split(
                ";"
            )  # [Subjects=0, Folders=0, Forms=0, Fields=0, LogLines=0]
            for part in parts:
                name, value = part.strip().split("=")
                # if value[-1] == ';':
                #     value = value[:-1]
                if name == "Subjects":
                    self.subjects_touched = int(value)
                elif name == "Folders":
                    self.folders_touched = int(value)
                elif name == "Forms":
                    self.forms_touched = int(value)
                elif name == "Fields":
                    self.fields_touched = int(value)
                elif name == "LogLines":
                    self.loglines_touched = int(value)
                else:
                    raise KeyError(
                        "Unknown Rave Object %s in response %s" % (name, success_stats)
                    )

        # Note: Metadata post has success_stats == 'N/A'
        self.new_records = r_get("NewRecords")


class RWSPostResponse(RWSResponse):
    """
    Parses responses from PostODMClinicalData messages with the format:

    .. code-block:: xml

        <Response ReferenceNumber="82e942b0-48e8-4cf4-b299-51e2b6a89a1b"
                  InboundODMFileOID=""
                  IsTransactionSuccessful="1"
                  SuccessStatistics="Rave objects touched: Subjects=0; Folders=0; Forms=0; Fields=0; LogLines=0" NewRecords=""
                  SubjectNumberInStudy="1103" SubjectNumberInStudySite="55">
        </Response>
    """

    def __init__(self, xml):
        # Call base class
        RWSResponse.__init__(self, xml)
        r_get = self.root.get
        # These counts may only come as a result of a Clinical data POST
        snis = r_get("SubjectNumberInStudy")
        self.subjects_in_study = int(snis) if snis is not None else None

        sniss = r_get("SubjectNumberInStudySite")
        self.subjects_in_study_site = int(sniss) if sniss is not None else None

        # DraftImported only comes from a MetaData Post
        # In which case successStatistics will be SuccessStatistics="N/A"
        self.draft_imported = r_get("DraftImported")


class RWSPostErrorResponse(RWSErrorResponse):
    """
    Responses to Clinical data post messages have additional Attributes to normal RWS Response messages:

    .. code-block:: xml

        <Response
            ReferenceNumber="5b1fa9a3-0cf3-46b6-8304-37c2e3b7d04f"
            InboundODMFileOID="1"
            IsTransactionSuccessful = "0"
            ReasonCode="RWS00024"
            ErrorOriginLocation="/ODM/ClinicalData[1]/SubjectData[1]"
            SuccessStatistics="Rave objects touched: Subjects=0; Folders=0; Forms=0; Fields=0; LogLines=0"
            ErrorClientResponseMessage="Subject already exists.">
        </Response>

    """

    def __init__(self, xml):
        """
        :param str xml: Error response
        """
        RWSErrorResponse.__init__(self, xml)

        # Get additional properties
        r_get = self.root.get
        self.reason_code = r_get("ReasonCode")
        self.error_origin_location = r_get("ErrorOriginLocation")
        self.error_client_response_message = r_get("ErrorClientResponseMessage")


class RWSStudyListItem(object):
    """An item in the RWS Study List response"""

    def __init__(
        self,
        oid=None,
        studyname=None,
        protocolname=None,
        environment=None,
        projecttype=None,
    ):
        """
        :param str oid: Study OID
        :param str studyname: Study Name
        :param str protocolname: Protocol Name
        :param str environment: Study Environment
        :param str projecttype: Project Type
        """
        self.oid = oid
        self.studyname = studyname
        self.protocolname = protocolname
        self.environment = environment
        self.projecttype = projecttype

    def isProd(self):
        """
        Is production if environment is empty

        :rtype: bool
        """
        return self.environment == "" and self.projecttype != "GlobalLibraryVolume"

    @classmethod
    def fromElement(cls, elem):
        """
        Read properties from an XML Element to build a StudyList Item

        :param lxml.etree.Element elem: The source Study XML Element

        .. code-block:: xml

            <Study OID="Fixitol(Dev)" mdsol:ProjectType="GlobalLibraryVolume">
                <GlobalVariables>
                    <StudyName>Fixitol (Dev)</StudyName>
                    <StudyDescription/>
                    <ProtocolName>Fixitol</ProtocolName>
                </GlobalVariables>
            </Study>

        """
        e_global_variables = elem.find(ODM_NS + "GlobalVariables")
        studyname = e_global_variables.find(ODM_NS + "StudyName").text
        protocolname = e_global_variables.find(ODM_NS + "ProtocolName").text
        self = cls(
            oid=elem.get("OID"),
            projecttype=elem.get(MEDI_NS + "ProjectType", "Project"),
            studyname=studyname,
            protocolname=protocolname,
            environment=getEnvironmentFromNameAndProtocol(studyname, protocolname),
        )

        return self


# I hate multi-inheritance generally but since this is inheriting from a built-in like list I feel
# less bad about it.
class RWSStudies(list, ODMDoc):
    """
    Represents a list of studies. Extends the list class and adds a couple of extra properties:

    .. code-block:: xml

        <ODM FileType="Snapshot" FileOID="767a1f8b-7b72-4d12-adbe-37d4d62ba75e"
             CreationDateTime="2013-04-08T10:02:17.781-00:00"
             ODMVersion="1.3"
             xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata"
             xmlns:xlink="http://www.w3.org/1999/xlink"
             xmlns="http://www.cdisc.org/ns/odm/v1.3">
             <Study OID="Fixitol(Dev)">
                <GlobalVariables>
                      <StudyName>Fixitol (Dev)</StudyName>
                      <StudyDescription/>
                      <ProtocolName>Fixitol</ProtocolName>
                </GlobalVariables>
             </Study>
             <Study OID="IANTEST(Prod)">
                <GlobalVariables>
                      <StudyName>IANTEST</StudyName>
                      <StudyDescription/>
                      <ProtocolName>IANTEST</ProtocolName>
                </GlobalVariables>
             </Study>
        </ODM>
    """

    def __init__(self, xml):
        # Get basic properties
        ODMDoc.__init__(self, xml)

        for estudy in self.root.findall(ODM_NS + "Study"):
            self.append(RWSStudyListItem.fromElement(estudy))


class MetaDataVersion(object):
    """
    A single MetaDataVersion instance

    .. code-block:: xml

        <MetaDataVersion OID="1203" Name="Webservice Outbound"/>
    """

    def __init__(self):
        self.oid = None
        self.name = None

    @classmethod
    def fromElement(cls, elem):
        """
        Read properties from a MetaDataVersion element

        :param lxml.etree._Element elem: Source etree Element
        """

        self = cls()

        self.oid = elem.get("OID")
        self.name = elem.get("Name")
        return self


class RWSStudyMetadataVersions(list, ODMDoc, RWSStudyListItem):
    """
    Parses responses from MetaDataVersions request:

    .. code-block:: xml

        <ODM ODMVersion="1.3" Granularity="Metadata" FileType="Snapshot" FileOID="d26b4d33-376d-4037-9747-684411190179" CreationDateTime=" 2013-04-08T01:29:13 " xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata">
            <Study OID="IANTEST">
                <GlobalVariables>
                    <StudyName>IANTEST</StudyName>
                    <StudyDescription></StudyDescription>
                    <ProtocolName>IANTEST</ProtocolName>
                </GlobalVariables>
                <MetaDataVersion OID="1203" Name="Webservice Outbound" />
                <MetaDataVersion OID="1195" Name="JC_Demo_Draft1" />
                <MetaDataVersion OID="1165" Name="Initial" />
            </Study>
        </ODM>
    """

    def __init__(self, xml):
        # Get basic properties
        ODMDoc.__init__(self, xml)

        root = self.root  # type: lxml.etree._Element

        e_study = root.find(ODM_NS + "Study")  # type: lxml.etree._Element

        # Quick way to grab the elements here, nasty though
        self.study = RWSStudyListItem.fromElement(e_study)

        for e_version in e_study.findall(ODM_NS + "MetaDataVersion"):
            self.append(MetaDataVersion.fromElement(e_version))


class RWSSubjectListItem(object):
    """
    Parses response of Subject List request:

    .. code-block:: xml

        <ClinicalData StudyOID="Fixitol(Dev)" MetaDataVersionOID="1111">
         <SubjectData SubjectKey="1111">
            <SiteRef LocationOID="335566"/>
         </SubjectData>
        </ClinicalData>

    Optionally ClinicalData may include status:

    .. code-block:: xml

        <SubjectData SubjectKey="1111" mdsol:Overdue="No"
                  mdsol:Touched="Yes"
                  mdsol:Empty="No"
                  mdsol:Incomplete="No"
                  mdsol:NonConformant="No"
                  mdsol:RequiresSecondPass="No"
                  mdsol:RequiresReconciliation="No"
                  mdsol:RequiresVerification="No"
                  mdsol:Verified="No"
                  mdsol:Frozen="No"
                  mdsol:Locked="No"
                  mdsol:RequiresReview="No"
                  mdsol:PendingReview="No"
                  mdsol:Reviewed="No"
                  mdsol:RequiresAnswerQuery="No"
                  mdsol:RequiresPendingCloseQuery="No"
                  mdsol:RequiresCloseQuery="No"
                  mdsol:StickyPlaced="No"
                  mdsol:Signed="No"
                  mdsol:SignatureCurrent="No"
                  mdsol:RequiresTranslation="No"
                  mdsol:RequiresCoding="No"
                  mdsol:RequiresPendingAnswerQuery="No"
                  mdsol:RequiresSignature="No"
                  mdsol:ReadyForFreeze="No"
                  mdsol:ReadyForLock="Yes">

    The SubjectKey can be either a Subject ID or a UUID depending on the value of SubjectKeyType:

    .. code-block:: xml

        <ClinicalData StudyOID="Fixitol(Dev)" MetaDataVersionOID="1111">
         <SubjectData SubjectKey="EC82F1AB-D463-4930-841D-36FC865E63B2" mdsol:SubjectName="1" mdsol:SubjectKeyType="SubjectUUID">
            <SiteRef LocationOID="335566"/>
         </SubjectData>
        </ClinicalData>

    The Response may include links:

    .. code-block:: xml

        <ClinicalData StudyOID="Fixitol(Dev)" MetaDataVersionOID="1111">
         <SubjectData SubjectKey="1111">
            <SiteRef LocationOID="335566"/>
            <mdsol:Link xlink:type="simple" xlink:href="http://innovate.mdsol.com/MedidataRAVE/HandleLink.aspx?page=SubjectPage.aspx?ID=849" />
         </SubjectData>
        </ClinicalData>

    """

    STATUS_PROPERTIES = [
        "Overdue",
        "Touched",
        "Empty",
        "Incomplete",
        "NonConformant",
        "RequiresSecondPass",
        "RequiresReconciliation",
        "RequiresVerification",
        "Verified",
        "Frozen",
        "Locked",
        "RequiresReview",
        "PendingReview",
        "Reviewed",
        "RequiresAnswerQuery",
        "RequiresPendingCloseQuery",
        "RequiresCloseQuery",
        "StickyPlaced",
        "Signed",
        "SignatureCurrent",
        "RequiresTranslation",
        "RequiresCoding",
        "RequiresPendingAnswerQuery",
        "RequiresSignature",
        "ReadyForFreeze",
        "ReadyForLock",
        "SubjectKeyType",
        "SubjectName",
    ]

    def __init__(self):
        """The ODM message has a ClinicalData element with a single SubjectData and SiteRef elements
           nested within. I collapse into a single object
        """
        self.studyoid = None
        self.metadataversionoid = None
        self.subjectkey = None
        self.subjectkeytype = None
        self.locationoid = None
        self.study_environment_site_number = None

        self.active = None  # SubjectActive
        self.deleted = None  # Deleted
        self.links = []  # Link if requested

        # Optional properties, only if status included
        for prop in RWSSubjectListItem.STATUS_PROPERTIES:
            setattr(self, prop.lower(), None)

    @property
    def subject_name(self):
        """
        Get the subject name consistently
        :rtype str
        :return: The Subject ID for the subject

        .. note::
          * If the `SubjectKeyType` is `SubjectUUID` then the subject name lives in the `mdsol:SubjectName` attribute
          * If the `SubjectKeyType` is `SubjectName` then the subject name lives in the `SubjectKey` attribute
        """
        if self.subjectkeytype and self.subjectkeytype == "SubjectUUID".lower():
            # if the SubjectKeyType is "SubjectUUID", then return the SubjectName
            return self.subjectname
        else:
            return self.subjectkey

    @classmethod
    def fromElement(cls, elem):
        """
        Read properties from an XML Element
        """
        self = cls()
        self.studyoid = elem.get("StudyOID")
        self.metadataversionoid = elem.get("MetaDataVersionOID")
        e_subjectdata = elem.findall(ODM_NS + "SubjectData")[0]
        self.subjectkey = e_subjectdata.get("SubjectKey")
        e_siteref = e_subjectdata.findall(ODM_NS + "SiteRef")[0]
        self.locationoid = e_siteref.get("LocationOID")
        self.study_environment_site_number = e_siteref.get(MEDI_NS + "StudyEnvSiteNumber")

        e_links = e_subjectdata.findall(MEDI_NS + "Link")
        for e_link in e_links:
            self.links.append(e_link.get(XLINK_NS + "href"))

        decodes = {"yes": True, "no": False, "": None}
        for prop in RWSSubjectListItem.STATUS_PROPERTIES:
            val = e_subjectdata.get(MEDI_NS + prop, "").lower()
            setattr(self, prop.lower(), decodes.get(val, val))

        # By default we only get back active and non-deleted subjects
        self.active = decodes[
            e_subjectdata.get(MEDI_NS + "SubjectActive", "yes").lower()
        ]
        self.deleted = decodes[e_subjectdata.get(MEDI_NS + "Deleted", "no").lower()]

        return self


class RWSSubjects(list, ODMDoc):
    """
    Represents a list of subjects:

    .. code-block:: xml

        <ODM FileType="Snapshot"
             FileOID="770f1758-db33-4ab2-af72-38db863734aa"
             CreationDateTime="2013-04-08T14:08:06.875-00:00"
             ODMVersion="1.3">

             <ClinicalData StudyOID="Fixitol(Dev)" MetaDataVersionOID="1111">
                <SubjectData SubjectKey="000001">
                   <SiteRef LocationOID="BP001"/>
                   <mdsol:Link xlink:type="simple" xlink:href="http://innovate.mdsol.com/MedidataRAVE/HandleLink.aspx?page=SubjectPage.aspx?ID=849" />
                </SubjectData>
             </ClinicalData>

             <ClinicalData StudyOID="Fixitol(Dev)" MetaDataVersionOID="1111">
                 <SubjectData SubjectKey="1111">
                    <SiteRef LocationOID="335566"/>
                    <mdsol:Link xlink:type="simple" xlink:href="http://innovate.mdsol.com/MedidataRAVE/HandleLink.aspx?page=SubjectPage.aspx?ID=849" />
                 </SubjectData>
             </ClinicalData>
        </ODM>

    """

    def __init__(self, xml):
        # Get basic properties
        ODMDoc.__init__(self, xml)

        root = self.root  # from ODMDoc

        for e_clindata in root.findall(ODM_NS + "ClinicalData"):
            self.append(RWSSubjectListItem.fromElement(e_clindata))



# -*- coding: utf-8 -*-
import datetime
import sys

from mock import patch

from ..common import obj_to_doc

__author__ = 'isparks'

import unittest

from rwslib.builders.common import bool_to_yes_no, bool_to_true_false, ODMElement, get_utc_date
from rwslib.builders.clinicaldata import UserRef, LocationRef, ClinicalData, SubjectData
from rwslib.builders.metadata import Study
from rwslib.builders.admindata import AdminData
from rwslib.builders.core import ODM
from xml.etree import cElementTree as ET


class TestBoolToTrueFalse(unittest.TestCase):
    def test_true_to_TRUE(self):
        """TRUE returned from true"""
        self.assertEqual('TRUE', bool_to_true_false(True))

    def test_false_to_FALSE(self):
        """FALSE returned from false"""
        self.assertEqual('FALSE', bool_to_true_false(False))


class TestBoolToYesNo(unittest.TestCase):
    def test_true_to_TRUE(self):
        """TRUE returned from true"""
        self.assertEqual('Yes', bool_to_yes_no(True))

    def test_false_to_FALSE(self):
        """FALSE returned from false"""
        self.assertEqual('No', bool_to_yes_no(False))


class TestInheritance(unittest.TestCase):
    """The things we do for 100% coverage."""

    def test_inheritance_warning(self):
        class NewObj(ODMElement):
            """We do not override the __lshift__ method"""
            pass

        with self.assertRaises(ValueError):
            # Exercise __lshift__
            NewObj() << object()


class TestString(unittest.TestCase):
    def test_to_string(self):
        self.assertEqual('<UserRef UserOID="test" />', str(UserRef("test")))


class TestAttributeSetters(unittest.TestCase):
    class TestElem(ODMElement):
        """Test class with a bad __lshift__ implementation"""

        def __init__(self):
            self.user = None
            self.locations = []

        def __lshift__(self, other):
            self.set_single_attribute(other, UserRef, "xxxuser")  # Incorrect spelling of user attribute
            self.set_list_attribute(other, LocationRef, "xxxlocations")  # Incorrect spelling of location attribute

    def test_single_attribute_misspelling(self):
        tested = TestAttributeSetters.TestElem()
        with self.assertRaises(AttributeError):
            tested << UserRef("Fred")

    def test_list_attribute_misspelling(self):
        tested = TestAttributeSetters.TestElem()
        with self.assertRaises(AttributeError):
            tested << LocationRef("Site 22")


class TestODM(unittest.TestCase):

    def test_valid_children(self):
        """We can add Study, AdminData or ClinicalData"""
        obj = ODM("Test User", fileoid="1234", description="Some Great Study")
        obj << Study(oid="Study1")
        obj << AdminData(study_oid="Study1")
        obj << ClinicalData(projectname="Study1", environment="Prod")
        tested = ET.fromstring(str(obj))
        self.assertEqual("{http://www.cdisc.org/ns/odm/v1.3}ODM", tested.tag)
        self.assertEqual("Transactional", tested.get('FileType'))
        self.assertEqual("Some Great Study", tested.get('Description'))
        self.assertEqual(3, len(list(tested)))

    def test_invalid_children(self):
        """We can add Study, AdminData or ClinicalData"""
        obj = ODM("Test User", fileoid="1234")
        with self.assertRaises(ValueError) as exc:
            obj << SubjectData("Site1", "Subject 1")
        self.assertEqual("ODM object can only receive ClinicalData, Study or AdminData object", str(exc.exception))

    def test_valid_granularity(self):
        """We can add Study, AdminData or ClinicalData"""
        with self.assertRaises(ValueError) as exc:
            obj = ODM("Test User", fileoid="1234",
                      description="Some Great Study", granularity="NomNom")
        if (sys.version_info > (3, 0)):
            self.assertEqual("Should be an instance of GranularityType not <class 'str'>", str(exc.exception))
        else:
            self.assertEqual("Should be an instance of GranularityType not <type 'str'>", str(exc.exception))

    def test_source_system(self):
        """We can add a SourceSystem, SourceSystemVersion """
        obj = ODM("Test User", fileoid="1234", source_system="Battlestar", source_system_version="1.04")
        tested = obj_to_doc(obj=obj)
        self.assertEqual("Battlestar", tested.get('SourceSystem'))
        self.assertEqual("1.04", tested.get('SourceSystemVersion'))

    def test_creation_datetime(self):
        """Create multiple ODM and get different datetimes"""
        obj_1 = ODM("Test User", fileoid="1234", source_system="Battlestar", source_system_version="1.04")
        tested_1 = obj_to_doc(obj=obj_1)
        with patch('rwslib.builders.common.datetime') as mock_dt:
            # offset the time to ensure we don't overlap
            mock_dt.utcnow.return_value = get_utc_date() + datetime.timedelta(seconds=61)
            obj_2 = ODM("Test User", fileoid="1235", source_system="Battlestar", source_system_version="1.04")
            tested_2 = obj_to_doc(obj=obj_2)
        self.assertEqual(tested_1.get('Originator'), tested_2.get('Originator'))
        self.assertEqual(tested_1.get('SourceSystem'), tested_2.get('SourceSystem'))
        self.assertNotEqual(tested_1.get('FileOID'), tested_2.get('FileOID'))
        self.assertNotEqual(tested_1.get('CreationDateTime'), tested_2.get('CreationDateTime'))

    def test_respect_unique_file_oid(self):
        """Test the RespectUniqueFileOID attribute"""
        obj_1 = ODM("Test User", fileoid="1234", source_system="Battlestar", source_system_version="1.04", respect_unique_file_oid=True)
        tested_1 = obj_to_doc(obj=obj_1)
        self.assertEqual(tested_1.get('RespectUniqueFileOID'), "Yes")
        obj_2 = ODM("Test User", fileoid="1234", source_system="Battlestar", source_system_version="1.04", respect_unique_file_oid=False)
        tested_2 = obj_to_doc(obj=obj_2)
        self.assertEqual(tested_2.get('RespectUniqueFileOID'), "No")



# -*- coding: utf-8 -*-
__author__ = 'anewbigging'

import unittest


def all_tests():
    suite = unittest.TestSuite()
    loader = unittest.TestLoader()
    suite.addTests(loader.discover('rwscmd.tests'))
    return suite


from rwslib import RWSConnection
from rwslib.rws_requests import *
from rwslib.rws_requests.odm_adapter import *
from rwslib.rws_requests.biostats_gateway import *

if __name__ == '__main__':

    from _settings import accounts

    # Accounts is a dict of dicts like
    # accounts = {'innovate' : {'username': 'username',
    #                          'password':'password'},
    #             'otherurl' : {'username': 'username',
    #                          'password':'password'},
    #            }

    acc = accounts['innovate']
    rave = RWSConnection('innovate', acc['username'], acc['password'])

    print(rave.send_request(VersionRequest(), retries=3))

    print(rave.send_request(BuildVersionRequest()))
    print(rave.send_request(DiagnosticsRequest()))
    print(rave.send_request(CacheFlushRequest()).istransactionsuccessful)

    print("Clinical studies request")
    studies = rave.send_request(ClinicalStudiesRequest(), timeout=60)
    print(len(studies))
    print(rave.last_result.url)
    print(rave.last_result.text)
    print(rave.request_time)

    print("Metadata studies request")
    m_studies = rave.send_request(MetadataStudiesRequest())
    print(len(m_studies))
    print(rave.last_result.text)

    print("Study %s Drafts" % 'Mediflex')
    drafts = rave.send_request(StudyDraftsRequest('Mediflex'))
    for draft in drafts:
        print(draft.oid.ljust(20), draft.name)

    print("Study %s Versions" % 'Mediflex')
    versions = rave.send_request(StudyVersionsRequest('Mediflex'))
    for version in versions:
        print(version.oid.ljust(20), version.name)

    sv1 = versions[0]
    version_odm = rave.send_request(StudyVersionRequest('Mediflex', sv1.oid))
    print("Mediflex Study version OID = %s %s" % (str(sv1.oid), version_odm[0:50] + '...',))

    gl_studies = rave.send_request(GlobalLibrariesRequest())
    print("There are %d global libaries" % len(gl_studies))
    for study in gl_studies:
        print(study.studyname)
    gl_1 = gl_studies[0]

    gl_drafts = rave.send_request(GlobalLibraryDraftsRequest(gl_1.studyname))
    print("There are %d drafts for GlobalLibrary %s" % (len(gl_drafts), gl_1.studyname,))

    gl_versions = rave.send_request(GlobalLibraryVersionsRequest(gl_1.studyname))
    print("There are %d versions for GlobalLibrary %s" % (len(gl_versions), gl_1.studyname,))

    glv1 = gl_versions[0]
    gl1_version_odm = rave.send_request(GlobalLibraryVersionRequest(gl_1.studyname, glv1.oid))
    print("Global library 1, OID = %s %s" % (str(glv1.oid), gl1_version_odm[0:50] + '...',))

    try:
        subs = rave.send_request(StudySubjectsRequest('SimpleStudy', 'PROD', status=True))
        print(rave.last_result.url)
    except:
        print(rave.last_result.url)
        raise
    print("SimpleStudy(PROD) has %d subjects " % len(subs))

    try:
        subs = rave.send_request(StudySubjectsRequest('SimpleStudy', 'TEST'))
    except:
        print(rave.last_result.url)
        raise
    print("SimpleStudy(TEST) has %d subjects " % len(subs))

    data = u"""<?xml version="1.0" encoding="utf-8" ?>
<ODM CreationDateTime="2013-06-17T17:03:29" FileOID="3b9fea8b-e825-4e5f-bdc8-1464bdd7a664" FileType="Transactional" ODMVersion="1.3" xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata">
  <ClinicalData MetaDataVersionOID="DEFAULT" StudyOID="SIMPLESTUDY (TEST)">
    <SubjectData SubjectKey="012" TransactionType="Insert">
      <SiteRef LocationOID="TESTSITE" />
      <StudyEventData StudyEventOID="SUBJECT" TransactionType="Update">
        <FormData FormOID="ENROL">
          <ItemGroupData ItemGroupOID="ENROL">
            <ItemData ItemOID="SUBID" Value="012" />
            <ItemData ItemOID="BIRTHDT" Value="1973 Jun 26" />
          </ItemGroupData>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>
</ODM>"""

    # Post data example
    res = rave.send_request(PostDataRequest(data))
    print(res)

    data = """<ODM FileType="Snapshot" Granularity="Metadata" CreationDateTime="2013-06-18T15:09:54.843-00:00" FileOID="82370e27-a6a5-41dc-8c07-829e489823df" ODMVersion="1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" xmlns="http://www.cdisc.org/ns/odm/v1.3">
  <Study OID="SIMPLESTUDY">
    <GlobalVariables>
      <StudyName>SIMPLESTUDY</StudyName>
      <StudyDescription></StudyDescription>
      <ProtocolName>SIMPLESTUDY</ProtocolName>
    </GlobalVariables>
    <BasicDefinitions/>
    <MetaDataVersion OID="1" Name="Version 5" mdsol:PrimaryFormOID="ENROL" mdsol:DefaultMatrixOID="DEFAULT" mdsol:SignaturePrompt="Sign this please.">
      <Protocol>
        <StudyEventRef StudyEventOID="SCREEN" OrderNumber="1" Mandatory="No" />
      </Protocol>
      <StudyEventDef OID="SCREEN" Name="Screening" Type="Common" Repeating="Yes" mdsol:OrderNumber="1">
        <FormRef FormOID="VITAL" OrderNumber="2" Mandatory="No" />
      </StudyEventDef>
      <FormDef OID="ENROL" Repeating="No" Name="Enrol" mdsol:OrderNumber="1" mdsol:ConfirmationStyle="None">
        <ItemGroupRef ItemGroupOID="ENROL" Mandatory="Yes" />
      </FormDef>
      <FormDef OID="VITAL" Repeating="No" Name="Vitals" mdsol:OrderNumber="2">
        <ItemGroupRef ItemGroupOID="VITAL" Mandatory="Yes" />
      </FormDef>
      <ItemGroupDef OID="ENROL" Name="ENROL" Repeating="No">
        <ItemRef ItemOID="SUBID" OrderNumber="1" Mandatory="No" />
        <ItemRef ItemOID="BIRTHDT" OrderNumber="2" Mandatory="No" />
      </ItemGroupDef>
      <ItemGroupDef OID="VITAL" Name="VITAL" Repeating="No">
        <ItemRef ItemOID="VDAT" OrderNumber="1" Mandatory="No" />
        <ItemRef ItemOID="WEIGHT_KG" OrderNumber="2" Mandatory="No" />
        <ItemRef ItemOID="HEIGHT_CM" OrderNumber="3" Mandatory="No" />
      </ItemGroupDef>
      <ItemDef OID="SUBID" Name="SUBID" DataType="text" Length="10" mdsol:ControlType="Text">
        <Question>
          <TranslatedText xml:lang="en">Subject ID</TranslatedText>
        </Question>
      </ItemDef>

      <ItemDef OID="VDAT" Name="VDAT" DataType="date" mdsol:DateTimeFormat="yyyy MMM dd" mdsol:ControlType="DateTime">
        <Question>
          <TranslatedText xml:lang="en">Visit Date</TranslatedText>
        </Question>
      </ItemDef>
      <ItemDef OID="BIRTHDT" Name="BIRTHDT" DataType="date" mdsol:DateTimeFormat="yyyy MMM dd" mdsol:ControlType="DateTime">
        <Question>
          <TranslatedText xml:lang="en">Date of Birth</TranslatedText>
        </Question>
      </ItemDef>
      <ItemDef OID="WEIGHT_KG" Name="WEIGHT_KG" DataType="float" Length="4" SignificantDigits="1" mdsol:ControlType="Text">
        <Question>
          <TranslatedText xml:lang="en">Weight</TranslatedText>
        </Question>
      </ItemDef>
      <ItemDef OID="HEIGHT_CM" Name="HEIGHT_CM" DataType="float" Length="4" SignificantDigits="1" mdsol:ControlType="Text">
        <Question>
          <TranslatedText xml:lang="en">Height</TranslatedText>
        </Question>
      </ItemDef>
    </MetaDataVersion>
  </Study>
</ODM>"""

    try:
        res = rave.send_request(PostMetadataRequest("SIMPLESTUDY", data))
    except:
        print(rave.last_result.url)
        raise
    print(res)

    print("Study")
    study = rave.send_request(StudyDatasetRequest('SIMPLESTUDY', 'TEST', formoid="VITAL"))
    print(study)

    print("Version")
    version = rave.send_request(VersionDatasetRequest('SIMPLESTUDY', 'TEST', 1128, formoid="VITAL"))
    print(version)

    print("Subject")
    subject = rave.send_request(SubjectDatasetRequest('SIMPLESTUDY', 'TEST', '1', formoid="VITAL"))
    print(subject)

    cv_metadata_odm = rave.send_request(CVMetaDataRequest('SIMPLESTUDY', 'TEST', rawsuffix='RAW'))
    print("CV Metadata")
    print(cv_metadata_odm)

    print("CSV Data for VITAL")
    vital_csv_data = rave.send_request(FormDataRequest('SIMPLESTUDY', 'TEST', 'REGULAR', 'VITAL', dataset_format="csv"))
    print(vital_csv_data)

    print("All CSV Metadata")
    all_csv_meta = rave.send_request(MetaDataRequest(dataset_format='csv'))
    print(all_csv_meta)

    print("SIMPLESTUDY Project Metadata")
    proj_csv_meta = rave.send_request(ProjectMetaDataRequest('SIMPLESTUDY'))
    print(proj_csv_meta)

    print("SIMPLESTUDY VITALS View Metadata")
    view_csv_meta = rave.send_request(ViewMetaDataRequest("V_SIMPLESTUDY_VITAL", dataset_format='xml'))
    print(view_csv_meta)

    print("SIMPLESTUDY(TEST) Comments")
    com_csv = rave.send_request(CommentDataRequest("SIMPLESTUDY", "TEST"))
    print(com_csv)

    print("SIMPLESTUDY(TEST) Deviations")
    dev_csv = rave.send_request(ProtocolDeviationsRequest("SIMPLESTUDY", "TEST"))
    print(dev_csv)

    print("CDASH Forms(PROD) Data Dictionaries")
    dd_csv = rave.send_request(DataDictionariesRequest("CDASH Forms", "PROD", dataset_format="xml"))
    print(dd_csv)

    print("First 100 rows of Audits for SIMPLESTUDY TEST")
    audits = rave.send_request(AuditRecordsRequest('SIMPLESTUDY', 'TEST'))
    print(rave.last_result.headers)  # Get headers, next and last entries?
    print(audits)

    print("Folder Definitions for SIMPLESTUDY")
    version_folders = rave.send_request(VersionFoldersRequest('SIMPLESTUDY', 'TEST'))
    print(version_folders)

    print("Signature Definitions for SIMPLESTUDY")
    sigdefs = rave.send_request(SignatureDefinitionsRequest('SIMPLESTUDY'))
    print(sigdefs)

    print("Site Definitions for SIMPLESTUDY TEST")
    sitedefs = rave.send_request(SitesMetadataRequest('SIMPLESTUDY', 'TEST'))
    print(sitedefs)

    print("Users for sites in SIMPLESTUDY TEST")
    users = rave.send_request(UsersRequest('SIMPLESTUDY', 'TEST'))
    print(users)
